# pylint:disable=unused-argument,arguments-differ
from __future__ import annotations
import angr.ailment as ailment

from angr.analyses.decompiler.sequence_walker import SequenceWalker
from angr.analyses.decompiler.structuring.structurer_nodes import (
    SequenceNode,
    CodeNode,
    MultiNode,
    LoopNode,
    ConditionNode,
    CascadingConditionNode,
)
from angr.analyses.decompiler.utils import is_empty_node


class CascadingIfsRemover(SequenceWalker):
    """
    Coalesce cascading If constructs. Transforming the following construct::

        if (cond_a) {
            if (cond_b) {
                true_body
            } else { }
        } else { }

    into::

        if (cond_a and cond_b) {
            true_body
        } else { }
    """

    def __init__(self, node):
        handlers = {
            SequenceNode: self._handle_Sequence,
            CodeNode: self._handle_Code,
            MultiNode: self._handle_MultiNode,
            LoopNode: self._handle_Loop,
            ConditionNode: self._handle_Condition,
            CascadingConditionNode: self._handle_CascadingCondition,
        }

        super().__init__(handlers)
        self.walk(node)

    def _handle_Condition(self, node, parent=None, index=None, **kwargs):
        """

        :param ConditionNode node:
        :param successor:
        :return:
        """

        if node.true_node is not None:
            self._handle(node.true_node, parent=node, index=0)
        if node.false_node is not None:
            self._handle(node.false_node, parent=node, index=1)

        if node.true_node is not None and node.false_node is None:
            if isinstance(node.true_node, SequenceNode):
                last_node = None
                if len(node.true_node.nodes) > 1 and all(is_empty_node(node_) for node_ in node.true_node.nodes[:-1]):
                    last_node = node.true_node.nodes[-1]
                elif len(node.true_node.nodes) == 1:
                    last_node = node.true_node.nodes[0]

                true_node = last_node
            elif isinstance(node.true_node, ConditionNode):
                true_node = node.true_node
            else:
                return

            if (
                isinstance(true_node, ConditionNode)
                and true_node.true_node is not None
                and true_node.false_node is None
            ):
                node.condition = ailment.BinaryOp(
                    None, "LogicalAnd", (node.condition, true_node.condition), False, **node.condition.tags
                )
                node.true_node = true_node.true_node



from __future__ import annotations
import logging

import claripy
from claripy.ast.bv import BV
from pypcode import OpCode, Varnode, PcodeOp

from angr.engines.engine import SimEngine
from angr.utils.constants import DEFAULT_STATEMENT
from .lifter import IRSB
from .behavior import OpBehavior
from angr.errors import AngrError
from angr.state_plugins.inspect import BP_BEFORE, BP_AFTER

l = logging.getLogger(__name__)


class PcodeEmulatorMixin(SimEngine):
    """
    Mixin for p-code execution.
    """

    _current_op: PcodeOp | None
    _current_op_idx: int
    _current_behavior: OpBehavior | None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_op = None
        self._current_behavior = None

    def handle_pcode_block(self, irsb: IRSB) -> None:
        """
        Execute a single P-Code IRSB.

        :param irsb: Block to be executed.
        """
        self.irsb = irsb

        # Hack on a handler here to track whether exit has been handled or not
        # FIXME: Vex models this as a known exit statement, which we should also
        # do here. For now, handle it this way.
        self.state.scratch.exit_handled = False
        self._pcode_tmps = {}

        fallthru_addr = self.irsb.addr
        self.state.scratch.ins_addr = self.irsb.addr
        last_imark_op_idx = 0

        # Note: start_op_idx is instruction relative
        start_op_idx = self.state.scratch.statement_offset
        self.state.scratch.statement_offset = 0
        assert start_op_idx == 0, "FIXME: Test statement_offset behavior"

        for op_idx, op in enumerate(irsb._ops[start_op_idx:]):  # FIXME: Shouldn't use protected members of IRSB
            op_idx += start_op_idx

            if op.opcode == OpCode.IMARK:
                if op_idx > 0:
                    # Trigger BP for previous instruction once we reach next IMARK
                    self.state._inspect("instruction", BP_AFTER)

                decode_addr = op.inputs[0].offset
                last_imark_op_idx = op_idx

                # Note: instruction BP will not be triggered on p-code-relative jumps
                l.debug("Executing machine instruction @ %#x", decode_addr)
                for vn in op.inputs:
                    self.state._inspect("instruction", BP_BEFORE, instruction=vn.offset)

                # FIXME: Hacking this on here but ideally should use "scratch".
                self._pcode_tmps = {}  # FIXME: Consider alignment requirements
                self.state.scratch.ins_addr = decode_addr
                fallthru_addr = op.inputs[-1].offset + op.inputs[-1].size
                continue

            self._current_op = op
            self._current_op_idx = op_idx - last_imark_op_idx
            l.debug("Executing P-Code op: %s", self._current_op)
            self._execute_current_op()
            self._current_op = None

        if self.state.scratch.statement_offset == 0:
            self.state._inspect("instruction", BP_AFTER)

        if not self.state.scratch.exit_handled:
            self.successors.add_successor(
                self.state,
                fallthru_addr,
                self.state.scratch.guard,
                irsb.jumpkind,
                exit_stmt_idx=DEFAULT_STATEMENT,
                exit_ins_addr=self.state.scratch.ins_addr,
            )

    def _execute_current_op(self) -> None:
        """
        Execute the current p-code operation.
        """
        self._current_behavior = self.irsb.behaviors.get_behavior_for_opcode(self._current_op.opcode)

        if self._current_behavior.is_special:
            handlers = {
                OpCode.LOAD: self._execute_load,
                OpCode.STORE: self._execute_store,
                OpCode.BRANCH: self._execute_branch,
                OpCode.CBRANCH: self._execute_cbranch,
                OpCode.BRANCHIND: self._execute_branchind,
                OpCode.CALL: self._execute_call,
                OpCode.CALLIND: self._execute_callind,
                OpCode.CALLOTHER: self._execute_callother,
                OpCode.RETURN: self._execute_ret,
                OpCode.MULTIEQUAL: self._execute_multiequal,
                OpCode.INDIRECT: self._execute_indirect,
                OpCode.SEGMENTOP: self._execute_segment_op,
                OpCode.CPOOLREF: self._execute_cpool_ref,
                OpCode.NEW: self._execute_new,
            }
            handlers[self._current_behavior.opcode]()
        elif self._current_behavior.is_unary:
            self._execute_unary()
        else:
            self._execute_binary()

        self._current_behavior = None

    def _map_register_name(self, varnode: Varnode) -> int:
        """
        Map SLEIGH register offset to ArchInfo register offset based on name.

        :param varnode: Varnode to translate.
        :return:        Register file offset.
        """
        # FIXME: Will need performance optimization
        # FIXME: Should not get trans object this way. There should be a faster mapping method than going through trans
        reg_name = varnode.getRegisterName()
        try:
            reg_offset = self.state.project.arch.get_register_offset(reg_name.lower())
            l.debug("Mapped register '%s' to offset %x", reg_name, reg_offset)
        except ValueError:
            reg_offset = varnode.offset + 0x100000
            l.debug("Could not map register '%s' from archinfo. Mapping to %x", reg_name, reg_offset)
        return reg_offset

    @staticmethod
    def _adjust_value_size(num_bits: int, v_in: BV) -> BV:
        """
        Ensure given bv is num_bits bits long by either zero extending or truncating.
        """
        if v_in.size() > num_bits:
            v_out = v_in[num_bits - 1 : 0]
            l.debug("Truncating value %s (%d bits) to %s (%d bits)", v_in, v_in.size(), v_out, num_bits)
            return v_out
        if v_in.size() < num_bits:
            v_out = v_in.zero_extend(num_bits - v_in.size())
            l.debug("Extending value %s (%d bits) to %s (%d bits)", v_in, v_in.size(), v_out, num_bits)
            return v_out
        return v_in

    def _set_value(self, varnode: Varnode, value: BV) -> None:
        """
        Store a value for a given varnode.

        This method stores to the appropriate register, or unique space,
        depending on the space indicated by the varnode.

        :param varnode: Varnode to store into.
        :param value:   Value to store.
        """
        # FIXME: Consider moving into behavior.py
        value = self._adjust_value_size(varnode.size * 8, value)
        assert varnode.size * 8 == value.size()

        space = varnode.space
        l.debug("Storing %s %x %s %d", space.name, varnode.offset, value, varnode.size)
        if space.name == "register":
            self.state.registers.store(
                self._map_register_name(varnode), value, size=varnode.size, endness=self.project.arch.register_endness
            )

        elif space.name == "unique":
            self._pcode_tmps[varnode.offset] = value

        elif space.name.lower() in ("ram", "mem"):
            l.debug("Storing %s to offset %s", value, varnode.offset)
            self.state.memory.store(varnode.offset, value, endness=self.project.arch.memory_endness)

        else:
            raise AngrError(f"Attempted write to unhandled address space '{space.name}'")

    def _get_value(self, varnode: Varnode) -> BV:
        """
        Get a value for a given varnode.

        This method loads from the appropriate const, register, unique, or RAM
        space, depending on the space indicated by the varnode.

        :param varnode: Varnode to load from.
        :return:        Value loaded.
        """
        space_name = varnode.space.name
        size = varnode.size
        l.debug("Loading %s - %x x %d", space_name, varnode.offset, size)
        if space_name == "const":
            return claripy.BVV(varnode.offset, size * 8)
        if space_name == "register":
            return self.state.registers.load(
                self._map_register_name(varnode), size=size, endness=self.project.arch.register_endness
            )

        if space_name == "unique":
            # FIXME: Support loading data of different sizes. For now, assume
            # size of values read are same as size written.
            try:
                assert self._pcode_tmps[varnode.offset].size() == size * 8
            except KeyError:
                # FIXME: Add unique space to state tracking?
                l.warning("Uninitialized read from unique space offset %x", varnode.offset)
                self._pcode_tmps[varnode.offset] = claripy.BVV(0, size * 8)
            return self._pcode_tmps[varnode.offset]

        if space_name.lower() in ("ram", "mem"):
            val = self.state.memory.load(varnode.offset, endness=self.project.arch.memory_endness, size=size)
            l.debug("Loaded %s from offset %s", val, varnode.offset)
            return val

        raise AngrError(f"Attempted read from unhandled address space '{space_name}'")

    def _execute_unary(self) -> None:
        """
        Execute the unary behavior of the current op.
        """
        in0 = self._get_value(self._current_op.inputs[0])
        out = self._current_behavior.evaluate_unary(self._current_op.output.size, self._current_op.inputs[0].size, in0)
        self._set_value(self._current_op.output, out)

    def _execute_binary(self) -> None:
        """
        Execute the binary behavior of the current op.
        """

        # Validate output
        assert self._current_op.output is not None
        if (
            self._current_op.opcode
            in [
                OpCode.INT_LESS,
                OpCode.INT_SLESS,
                OpCode.INT_LESSEQUAL,
                OpCode.INT_SLESSEQUAL,
                OpCode.INT_EQUAL,
                OpCode.INT_NOTEQUAL,
            ]
            and self._current_op.output.size != 1
        ):
            l.warning(
                "SLEIGH spec states output size for op %s must be 1, but op has %d",
                self._current_op.opcode.__name__,
                self._current_op.output.size,
            )

        # Validate ops that mandate inputs of equal sizes
        # Validate ops that mandate output of greater size

        # Validate inputs

        in0 = self._get_value(self._current_op.inputs[0])
        in1 = self._get_value(self._current_op.inputs[1])
        out = self._current_behavior.evaluate_binary(
            self._current_op.output.size, self._current_op.inputs[0].size, in0, in1
        )
        self._set_value(self._current_op.output, out)

    def _execute_load(self) -> None:
        """
        Execute a p-code load operation.
        """
        space = self._current_op.inputs[0].getSpaceFromConst()
        offset = self._get_value(self._current_op.inputs[1])
        out = self._current_op.output
        if space.name.lower() in ("ram", "mem"):
            res = self.state.memory.load(offset, out.size, endness=self.project.arch.memory_endness)
        elif space.name in "register":
            res = self.state.registers.load(offset, size=out.size, endness=self.project.arch.register_endness)
        else:
            raise AngrError("Load from unhandled address space")
        l.debug("Loaded %s from offset %s", res, offset)
        self._set_value(out, res)

        # CHECKME: wordsize condition in cpuid load

    def _execute_store(self) -> None:
        """
        Execute a p-code store operation.
        """
        space = self._current_op.inputs[0].getSpaceFromConst()
        offset = self._get_value(self._current_op.inputs[1])
        data = self._get_value(self._current_op.inputs[2])
        l.debug("Storing %s at offset %s", data, offset)
        if space.name.lower() in ("ram", "mem"):
            self.state.memory.store(offset, data, endness=self.project.arch.memory_endness)
        elif space.name == "register":
            self.state.registers.store(offset, data, endness=self.project.arch.register_endness)
        else:
            raise AngrError("Store to unhandled address space")

    def _execute_branch(self) -> None:
        """
        Execute a p-code branch operation.
        """
        dest = self._current_op.inputs[0]
        if dest.space.name == "const":
            # P-Code-relative branch
            expr = self.state.scratch.ins_addr
            self.state.scratch.statement_offset = self._current_op_idx + dest.offset
        else:
            expr = dest.offset

        self.successors.add_successor(
            self.state,
            expr,
            self.state.scratch.guard,
            "Ijk_Boring",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

        self.state.scratch.exit_handled = True

    def _execute_cbranch(self) -> None:
        """
        Execute a p-code conditional branch operation.
        """
        exit_state = self.state.copy()
        cond = self._get_value(self._current_op.inputs[1])
        dest = self._current_op.inputs[0]

        if dest.space.name == "const":
            # P-Code-relative branch
            expr = exit_state.scratch.ins_addr
            exit_state.scratch.statement_offset = self._current_op_idx + dest.offset
        else:
            expr = dest.offset

        self.successors.add_successor(
            exit_state,
            expr,
            cond != 0,
            "Ijk_Boring",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

        cont_state = self.state
        cont_condition = cond == 0
        cont_state.add_constraints(cont_condition)
        cont_state.scratch.guard = claripy.And(cont_state.scratch.guard, cont_condition)

    def _execute_ret(self) -> None:
        """
        Execute a p-code return operation.
        """
        self.successors.add_successor(
            self.state,
            self._get_value(self._current_op.inputs[0]),
            self.state.scratch.guard,
            "Ijk_Ret",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

        self.state.scratch.exit_handled = True

    def _execute_branchind(self) -> None:
        """
        Execute a p-code indirect branch operation.
        """
        self.successors.add_successor(
            self.state,
            self._get_value(self._current_op.inputs[0]),
            self.state.scratch.guard,
            "Ijk_Boring",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

        self.state.scratch.exit_handled = True

    def _execute_call(self) -> None:
        """
        Execute a p-code call operation.
        """

        # FIXME: Spec claims CALL is semantically equivalent to BRANCH. But are p-code relative calls allowed? We assume
        #        not.

        self.successors.add_successor(
            self.state.copy(),  # FIXME: Check extra processing after call
            self._current_op.inputs[0].offset,
            self.state.scratch.guard,
            "Ijk_Call",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

        self.state.scratch.exit_handled = True

    def _execute_callind(self) -> None:
        """
        Execute a p-code indirect call operation.
        """
        self.successors.add_successor(
            self.state,
            self._get_value(self._current_op.inputs[0]),
            self.state.scratch.guard,
            "Ijk_Call",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

        self.state.scratch.exit_handled = True

    def _execute_callother(self) -> None:  # pylint:disable=no-self-use
        raise AngrError("CALLOTHER emulation not currently supported")

    def _execute_multiequal(self) -> None:  # pylint:disable=no-self-use
        raise AngrError("MULTIEQUAL appearing in unheritaged code?")

    def _execute_indirect(self) -> None:  # pylint:disable=no-self-use
        raise AngrError("INDIRECT appearing in unheritaged code?")

    def _execute_segment_op(self) -> None:  # pylint:disable=no-self-use
        raise AngrError("SEGMENTOP emulation not currently supported")

    def _execute_cpool_ref(self) -> None:  # pylint:disable=no-self-use
        raise AngrError("Cannot currently emulate cpool operator")

    def _execute_new(self) -> None:  # pylint:disable=no-self-use
        raise AngrError("Cannot currently emulate new operator")



from __future__ import annotations
import logging

from angr.sim_state import SimState
from .plugin import SimStatePlugin

l = logging.getLogger("angr.state_plugins.jni_references")


class SimStateJNIReferences(SimStatePlugin):
    """
    Management of the mapping between opaque JNI references and the
    corresponding Java objects.
    """

    def __init__(self, local_refs=None, global_refs=None):
        super().__init__()
        self.local_refs = local_refs or {}
        self.global_refs = global_refs or {}

    def lookup(self, opaque_ref):
        """
        Lookups the object that was used for creating the reference.
        """
        opaque_ref_value = self._get_reference_value(opaque_ref)
        # check local refs
        if opaque_ref_value in self.local_refs:
            return self.local_refs[opaque_ref_value]
        # check global refs
        if opaque_ref_value in self.global_refs:
            return self.global_refs[opaque_ref_value]
        raise KeyError(
            f"Unknown JNI reference {opaque_ref_value}. "
            f"Local references: {self.local_refs} "
            f"Global references: {self.global_refs}"
        )

    def create_new_reference(self, obj, global_ref=False):
        """
        Create a new reference thats maps to the given object.

        :param obj:              Object which gets referenced.
        :param bool global_ref:  Whether a local or global reference is created.
        """
        # get an unique address
        opaque_ref = self.state.project.loader.extern_object.allocate()
        # map the object to that address
        l.debug("Map %s to opaque reference 0x%x", obj, opaque_ref)
        if global_ref:
            self.global_refs[opaque_ref] = obj
        else:
            self.local_refs[opaque_ref] = obj
        return opaque_ref

    def clear_local_references(self):
        """
        Clear all local references.
        """
        self.local_refs = {}

    def delete_reference(self, opaque_ref, global_ref=False):
        """
        Delete the stored mapping of a reference.

        :param opaque_ref:       Reference which should be removed.
        :param bool global_ref:  Whether opaque_ref is a local or global
                                 reference.
        """
        opaque_ref_value = self._get_reference_value(opaque_ref)
        if global_ref:
            del self.global_refs[opaque_ref_value]
        else:
            del self.local_refs[opaque_ref_value]

    def _get_reference_value(self, opaque_ref):
        if self.state.solver.symbolic(opaque_ref):
            raise NotImplementedError(f"Opaque reference {opaque_ref.to_claripy()} is symbolic.")
        return self.state.solver.eval(opaque_ref)

    @SimStatePlugin.memo
    def copy(self, memo):  # pylint: disable=unused-argument
        return SimStateJNIReferences(local_refs=self.local_refs, global_refs=self.global_refs)

    def merge(self, others, merge_conditions, common_ancestor=None):  # pylint: disable=unused-argument
        l.warning("Merging is not implemented for JNI references!")
        return False

    def widen(self, others):  # pylint: disable=unused-argument
        l.warning("Widening is not implemented for JNI references!")
        return False


# TODO use a default JavaVM preset
#      see for reference: angr/engines/__init__.py
SimState.register_default("jni_references", SimStateJNIReferences)



from __future__ import annotations
from typing import NamedTuple, Any
from collections.abc import Callable
import multiprocessing
import platform
from angr.errors import AngrRuntimeError


class Closure(NamedTuple):
    """
    A pickle-able lambda; note that f, args, and kwargs must be pickleable
    """

    f: Callable[..., None]
    args: list[Any]
    kwargs: dict[str, Any]


class Initializer:
    """
    A singleton class with global state used to initialize a multiprocessing.Process
    """

    _single: Initializer | None = None

    @classmethod
    def get(cls) -> Initializer:
        """
        A wrapper around init since this class is a singleton
        """
        if cls._single is None:
            cls._single = cls(_manual=False)
        return cls._single

    def __init__(self, *, _manual: bool = True):
        if _manual:
            raise AngrRuntimeError("This is a singleton; call .get() instead")
        self.initializers: list[Closure] = []

    def register(self, f: Callable[..., None], *args: Any, **kwargs: Any) -> None:
        """
        A shortcut for adding Closures as initializers
        """
        self.initializers.append(Closure(f, args, kwargs))

    def initialize(self) -> None:
        """
        Initialize a multiprocessing.Process
        Set the current global initializer to the same state as this initializer, then calls each initializer
        """
        self._single = self
        for i in self.initializers:
            i.f(*i.args, **i.kwargs)


def mp_context():
    system = platform.system()
    spawn_methods = {
        "Windows": "spawn",
        "Linux": "fork",
        # Python<3.8 defaults to fork
        # https://bugs.python.org/issue33725
        "Darwin": "spawn",
    }
    spawn_method = spawn_methods.get(system, "fork")  # default to fork on other platforms
    return multiprocessing.get_context(spawn_method)



from __future__ import annotations
from collections import defaultdict
import logging
from typing import TypeVar

import networkx

from angr.ailment import Block, Address
from angr.ailment.statement import Assignment, ConditionalJump, Return, Jump
from angr.ailment.expression import Call, Const, BinaryOp, VirtualVariable, VirtualVariableCategory
from angr.analyses.s_liveness import SLivenessAnalysis
from angr.utils.ssa import is_phi_assignment
from angr.analyses import Analysis, AnalysesHub
from angr.analyses.s_reaching_definitions import SReachingDefinitionsAnalysis
from angr.knowledge_plugins.functions import Function
from angr.utils.graph import subgraph_between_nodes, Dominators, compute_dominance_frontier

_l = logging.getLogger(__name__)

T = TypeVar("T")


class Outliner(Analysis):
    """
    Outliner takes a function and some locations and attempts to outline the blocks within these locations into a
    separate function.
    """

    def __init__(
        self,
        func,
        ail_graph: networkx.DiGraph[Block],
        src_loc: Address,
        func_entry_loc: Address | None = None,
        frontier: set[Address] | None = None,
        vvar_id_start: int = 0xBEEF,
        block_addr_start: int = 0xAABB_0000,
        min_step: int = 1,
        liveness: SLivenessAnalysis | None = None,
    ):
        self.parent_func = func
        self.parent_graph = ail_graph
        self.vvar_id_start = vvar_id_start
        self.block_addr_start = block_addr_start
        self.min_step = min_step
        self.nodes_dict = {(node.addr, node.idx): node for node in self.parent_graph}

        self.src_loc = src_loc

        if func_entry_loc:
            self.parent_entry_loc = func_entry_loc
        else:
            func_entry_locs = [(bb.addr, bb.idx) for bb in self.parent_graph if self.parent_graph.in_degree[bb] == 0]
            if len(func_entry_locs) != 1:
                _l.warning("Graph has no obvious entry point")
            self.parent_entry_loc = min(func_entry_locs)

        self.parent_liveness = liveness or self.project.analyses[SLivenessAnalysis].prep()(
            self.parent_func,
            func_graph=self.parent_graph,
            entry=self.nodes_dict[self.parent_entry_loc],
            arg_vvars=[],  # TODO: FIXME
        )

        if frontier:
            self.frontier_locs = frontier
            self.frontier_vars = self._determine_frontier_vars()
        else:
            self.frontier_locs = self._determine_frontier_locs()
            self.frontier_vars = set()

        self.child_func, self.child_graph, self.child_funcargs = self._analyze()

    def _next_vvar_id(self) -> int:
        vvar_id = self.vvar_id_start
        self.vvar_id_start += 1
        return vvar_id

    def _next_block_addr(self) -> int:
        block_addr = self.block_addr_start
        self.block_addr_start += 1
        return block_addr

    def cleanup_callee_graph(self, g: networkx.DiGraph, func: Function):
        """
        Remove all phi assignments whose all source variables are undefined in the graph.
        """

        srda = self.project.analyses[SReachingDefinitionsAnalysis].prep()(func, func_graph=g).model

        to_kill = defaultdict(set)
        for phi_var_id, src_var_ids in srda.phivarid_to_varids.items():
            if all(srda.all_vvar_definitions[src_varid].is_extern for src_varid in src_var_ids):
                # remove the phi assignment
                phi_def_loc = srda.all_vvar_definitions[phi_var_id]
                assert phi_def_loc.block_addr is not None
                assert phi_def_loc.stmt_idx is not None
                phi_def_node = self.nodes_dict[(phi_def_loc.block_addr, phi_def_loc.block_idx)]
                to_kill[phi_def_node].add(phi_def_loc.stmt_idx)

        for node, kills in to_kill.items():
            node.statements = [stmt for i, stmt in enumerate(node.statements) if i not in kills]

    def get_interface(self, g: networkx.DiGraph[Block], func: Function) -> list[VirtualVariable]:
        """
        Recover the interface from a function AIL graph.
        """

        srda = self.project.analyses[SReachingDefinitionsAnalysis].prep()(func, func_graph=g).model

        blocks: dict[tuple[int, int | None], Block] = {(node.addr, node.idx): node for node in g}

        # find undefined vvars
        undef_vvars = []
        for vvar_id, defloc in srda.all_vvar_definitions.items():
            if defloc.is_extern:
                # remove undefined vvars that are only ever used in phi assignments
                use_locs = srda.all_vvar_uses[vvar_id]
                use_stmts = [
                    blocks[loc.addr, loc.block_idx].statements[loc.stmt_idx] for _, loc in use_locs if not loc.is_extern
                ]
                if not all(is_phi_assignment(stmt) for stmt in use_stmts):
                    undef_vvars.append(vvar_id)

        return [srda.varid_to_vvar[varid] for varid in undef_vvars]

    def _analyze(self):
        node_dict: dict[tuple[int, int | None], Block] = {
            (node.addr, node.idx): node for node in self.parent_graph.nodes
        }
        try:
            src_node = node_dict[self.src_loc]
        except KeyError as e:
            raise KeyError(f"Source location {self.src_loc} is not valid in the given graph.") from e

        frontier: list[Block] = []
        # ensure locs is valid
        for loc in self.frontier_locs:
            try:
                frontier.append(node_dict[loc])
            except KeyError as e:
                raise KeyError(f"Frontier location {loc} is not valid in the given graph.") from e

        # generate a subgraph
        subgraph = subgraph_between_nodes(
            self.parent_graph, src_node, frontier, include_frontier=False
        )  # FISHME: why was this True?

        in_edges = [(node, src_node) for node in self.parent_graph.pred[src_node]]
        out_edges = [
            (node, frontier_node)
            for frontier_node in frontier
            for node in self.parent_graph.pred[frontier_node]
            if node in subgraph and node not in frontier
        ]

        # remove the subgraph from the original graph
        for node in subgraph:
            self.parent_graph.remove_node(node)

        callee_func = Function(self.kb.functions, src_node.addr)
        callee_func.normalized = True
        # clean up the subgraph
        self.cleanup_callee_graph(subgraph, callee_func)
        # figure out the interface of the new callee
        callee_arg_vvars = self.get_interface(subgraph, callee_func)

        # rewrite the callsite
        vvar_id = self._next_vvar_id()
        call_expr = Call(
            None,
            f"outlined_func_{src_node.addr:x}",
            # Const(None, None, src_node.addr, 64),
            args=callee_arg_vvars,
            bits=self.project.arch.bits,
            ins_addr=src_node.addr,
        )
        switch_vvar = VirtualVariable(
            None, vvar_id, self.project.arch.bits, VirtualVariableCategory.REGISTER, oident=self.project.arch.ret_offset
        )
        call_stmt = Assignment(None, switch_vvar, call_expr, ins_addr=src_node.addr)
        new_src_node = Block(src_node.addr, src_node.original_size, [call_stmt], idx=src_node.idx)
        for pred, _ in in_edges:
            self.parent_graph.add_edge(pred, new_src_node)

        # build the return statement if needed
        if self.frontier_vars:
            srda = (
                self.project.analyses[SReachingDefinitionsAnalysis]
                .prep()(self.parent_func, func_graph=self.parent_graph)
                .model
            )
            ret_exprs = [srda.varid_to_vvar[idx] for idx in self.frontier_vars]
        else:
            ret_exprs = []

        retval_to_target: dict[int, tuple[int, int | None]] = {}
        if len(frontier) > 1:
            # there are multiple successors; this means the function must return to different locations. let's build
            # the dispatcher structure (at the return site in the caller) and the return nodes (in the callee)
            for succ in sorted(frontier, key=lambda node: (node.addr, node.idx)):
                retval_to_target[succ.addr] = succ.addr, succ.idx
            ret_exprs.append(switch_vvar)

        for ret_node, frontier_node in out_edges:
            if retval_to_target:
                new_ret_exprs = [*ret_exprs[:-1], Const(None, None, frontier_node.addr, self.project.arch.bits)]
            else:
                new_ret_exprs = ret_exprs
            ret_stmt = Return(None, new_ret_exprs, ins_addr=max(stmt.tags["ins_addr"] for stmt in ret_node.statements))

            ret_node_succs = list(subgraph.successors(ret_node))
            if len(ret_node_succs) == 0:
                # we can modify the ret_node directly
                if ret_node.statements and isinstance(ret_node.statements[-1], (ConditionalJump, Jump)):
                    del ret_node.statements[-1]
                ret_node.statements.append(ret_stmt)
            else:
                # we will have to create a new node and act as the successor of ret_node
                new_ret_node = Block(self._next_block_addr(), 0, [ret_stmt])
                if ret_node.statements and isinstance(ret_node.statements[-1], ConditionalJump):
                    cond_jump = ret_node.statements[-1]
                    if isinstance(cond_jump.true_target, Const) and cond_jump.true_target.value == frontier_node.addr:
                        _, cond_jump = cond_jump.replace(
                            cond_jump.true_target, Const(None, None, new_ret_node.addr, self.project.arch.bits)
                        )
                    if isinstance(cond_jump.false_target, Const) and cond_jump.false_target.value == frontier_node.addr:
                        _, cond_jump = cond_jump.replace(
                            cond_jump.false_target, Const(None, None, new_ret_node.addr, self.project.arch.bits)
                        )
                    ret_node.statements[-1] = cond_jump
                    subgraph.add_edge(ret_node, new_ret_node)

        if frontier:
            if retval_to_target:
                # build the dispatcher structure
                parent = new_src_node
                next_dispatcher_node_addr = self._next_block_addr(), None
                retval_to_target_items = sorted(retval_to_target.items())
                last_retval_to_target_item = retval_to_target_items.pop()
                for retval, jump_target in retval_to_target_items:
                    dispatcher_node_addr = next_dispatcher_node_addr
                    next_dispatcher_node_addr = self._next_block_addr(), None

                    retval_const = Const(None, None, retval, self.project.arch.bits)
                    cmp = BinaryOp(None, "CmpEQ", [switch_vvar, retval_const])
                    stmt = ConditionalJump(
                        None,
                        cmp,
                        Const(None, None, jump_target[0], self.project.arch.bits),
                        Const(None, None, next_dispatcher_node_addr[0], self.project.arch.bits),
                        true_target_idx=jump_target[1],
                        false_target_idx=next_dispatcher_node_addr[1],
                        ins_addr=dispatcher_node_addr[0],
                    )
                    dispatcher_node = Block(dispatcher_node_addr[0], 0, [stmt], dispatcher_node_addr[1])
                    self.parent_graph.add_edge(parent, dispatcher_node)
                    self.parent_graph.add_edge(dispatcher_node, node_dict[jump_target])

                    self._update_phi_stmts(node_dict[jump_target])

                    parent = dispatcher_node

                self.parent_graph.add_edge(parent, node_dict[last_retval_to_target_item[1]])

            else:
                (frontier_node,) = frontier
                self.parent_graph.add_edge(new_src_node, frontier_node)
                self._update_phi_stmts(frontier_node)

        if ret_exprs:
            if len(ret_exprs) > 1:
                _l.error("Outlined region seems to have multiple return values. Can't represent this correctly.")
            call_stmt.dst = ret_exprs[0]

        return callee_func, subgraph, callee_arg_vvars

    def _update_phi_stmts(self, block: Block):
        srcs = list(self.parent_graph.predecessors(block))
        src_addrs = [(src.addr, src.idx) for src in srcs]
        for stmt in block.statements:
            if is_phi_assignment(stmt):
                all_stmt_srcs = [src for src, _ in stmt.src.src_and_vvars]
                old_addrs = set(src_addrs) - set(all_stmt_srcs)
                new_addrs = set(all_stmt_srcs) - set(src_addrs)
                if len(old_addrs) == 1 and len(new_addrs) == 1:
                    # only source block is replaced by a new one
                    old_addr = next(iter(old_addrs))
                    new_addr = next(iter(new_addrs))
                    for idx, _ in enumerate(stmt.src.src_and_vvars):
                        src, vvars = stmt.src.src_and_vvars[idx]
                        if src == old_addr:
                            stmt.src.src_and_vvars[idx] = new_addr, vvars
                else:
                    # multiple source blocks have been replaced... it's bad
                    continue

    @staticmethod
    def _node_addr_to_str(addr: tuple[int, int | None]) -> str:
        """
        Convert a node address to a string representation.
        """
        return f"{addr[0]:#x}.{addr[1]}" if addr[1] is not None else f"{addr[0]:#x}"

    def _determine_frontier_vars(self) -> set[int]:
        """
        Given that the loc frontier has already been set, determine which variables are live when leaving that region.
        """

        return (
            set().union(*(self.parent_liveness.model.live_ins[f] for f in self.frontier_locs))
            - self.parent_liveness.model.live_ins[self.src_loc]
        )

    def _determine_frontier_locs(self) -> set[tuple[int, int | None]]:
        _l.debug("Determining the outlining frontier starting at (%#x, %s)", self.src_loc[0], self.src_loc[1])

        live_vars_dict = self.parent_liveness.live_vars_by_stmt()
        assert self.src_loc in live_vars_dict

        # find its dominance frontier
        doms = Dominators(self.parent_graph, self.nodes_dict[self.parent_entry_loc])
        dom_frontiers = compute_dominance_frontier(self.parent_graph, doms.dom)

        start = next(iter(bb for bb in self.parent_graph if (bb.addr, bb.idx) == self.src_loc))

        if start not in dom_frontiers:
            return set()

        queue = [start]
        frontiers = set()
        while queue:
            node = queue.pop(0)
            max_frontier = dom_frontiers[node]

            # for all nodes between `node` and frontier, see when new live variables are no longer live
            new_frontier = self._variable_life_frontier(node, max_frontier, min_step=self.min_step)
            _l.debug(
                "New frontier for node (%#x, %s): %s",
                node.addr,
                node.idx,
                [self._node_addr_to_str(x) for x in new_frontier],
            )

            frontiers |= new_frontier

        return frontiers

    def _variable_life_frontier(
        self, start: Block, max_frontier: set[Block], min_step=1
    ) -> set[tuple[int, int | None]]:
        """
        Find the frontier of blocks at which every variable defined in the start block dies.

        min_step will force the frontier to be at least that many steps away from the start block.
        """

        initial_live_vars = self.parent_liveness.model.live_ins[start.addr, start.idx]
        queue = [
            {
                "node": start,
                "live_vars": set(),
                "step": 0,
            }
        ]
        visited = {start}

        frontier = set()

        while queue:
            info = queue.pop(0)
            node = info["node"]
            step = info["step"]

            live_outs = self.parent_liveness.model.live_outs[node.addr, node.idx] - initial_live_vars
            _l.debug("Visiting node %#x[%s] (step %d). Live outs: %s", node.addr, node.idx, step, live_outs)
            if step >= min_step and not live_outs:
                frontier.add((node.addr, node.idx))
                continue

            for succ in self.parent_graph.successors(node):
                if succ in visited:
                    continue
                if succ in max_frontier:
                    frontier.add((succ.addr, succ.idx))
                    continue
                visited.add(succ)

                # continue to the next node
                queue.append(
                    {
                        "node": succ,
                        "live_vars": live_outs,
                        "step": step + 1,
                    }
                )

        return frontier


AnalysesHub.register_default("Outliner", Outliner)



from __future__ import annotations
from .amd64_ccalls import AMD64CCallRewriter
from .x86_ccalls import X86CCallRewriter
from .arm_ccalls import ARMCCallRewriter

CCALL_REWRITERS = {
    "X86": X86CCallRewriter,
    "AMD64": AMD64CCallRewriter,
    "ARMEL": ARMCCallRewriter,
}



# pylint:disable=unused-argument,import-outside-toplevel
from __future__ import annotations

import json
import logging
import os
import re
from typing import TypeVar, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pydantic_ai.models import Model
    from pydantic_ai.settings import ModelSettings
    from pydantic_ai.providers import Provider

T = TypeVar("T")

l = logging.getLogger(name=__name__)

Agent = None
OpenAIChatModel = None
OpenAIProvider = None


class LLMClient:
    """
    A client for interacting with LLMs via pydantic-ai.
    Used by the decompiler to suggest improved variable names, function names, and types.
    """

    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        api_base: str | None = None,
        max_tokens: int = 4096,
        temperature: float = 0.0,
    ):
        self.model = model
        self.api_key = api_key
        self.api_base = api_base
        self.max_tokens = max_tokens
        self.temperature = temperature
        self._pydantic_model: Model | str = self._build_model()

    def infer_provider(self, provider: str) -> Provider[Any]:
        """Infer the provider from the provider name."""
        if provider.startswith("gateway/"):
            from pydantic_ai.providers.gateway import gateway_provider  # type: ignore

            upstream_provider = provider.removeprefix("gateway/")
            return gateway_provider(upstream_provider)
        if provider in ("google-vertex", "google-gla"):
            from pydantic_ai.providers.google import GoogleProvider

            return GoogleProvider(vertexai=provider == "google-vertex")
        from pydantic_ai.providers import infer_provider_class

        provider_class = infer_provider_class(provider)
        return provider_class(api_key=self.api_key)  # type: ignore

    def _build_model(self) -> Model | str:
        """Build a pydantic-ai model object from the configured settings."""
        if self.api_base:
            global OpenAIChatModel, OpenAIProvider

            if OpenAIChatModel is None or OpenAIProvider is None:
                from pydantic_ai.models.openai import OpenAIChatModel  # type: ignore
                from pydantic_ai.providers.openai import OpenAIProvider  # type: ignore

            provider = OpenAIProvider(base_url=self.api_base, api_key=self.api_key or "no-key")
            return OpenAIChatModel(self.model, provider=provider)

        from pydantic_ai.models import infer_model

        model_name = self.model
        if ":" not in model_name:
            return self.model

        return infer_model(
            model_name,
            provider_factory=self.infer_provider,  # type:ignore
        )  # pylint:disable=unexpected-keyword-arg

    @classmethod
    def from_env(cls) -> LLMClient | None:
        """
        Create an LLMClient from environment variables.

        Uses ANGR_LLM_MODEL (required), ANGR_LLM_API_KEY (optional), and ANGR_LLM_API_BASE (optional).
        Returns None if ANGR_LLM_MODEL is not set.
        """
        model = os.environ.get("ANGR_LLM_MODEL")
        if not model:
            return None
        api_key = os.environ.get("ANGR_LLM_API_KEY")
        api_base = os.environ.get("ANGR_LLM_API_BASE")
        return cls(model=model, api_key=api_key, api_base=api_base)

    def _model_settings(self) -> ModelSettings:
        from pydantic_ai.settings import ModelSettings  # type:ignore

        return ModelSettings(temperature=self.temperature, max_tokens=self.max_tokens)

    def completion(self, messages: list[dict[str, str]], **kwargs) -> str:  # pylint:disable=unused-argument
        """
        Call the LLM with the given messages and return the response text.
        """

        global Agent

        if Agent is None:
            try:
                from pydantic_ai import Agent  # type: ignore
            except ImportError:
                raise ImportError(
                    "pydantic-ai is required for LLM support. You can install it with: pip install angr[llm] or "
                    "pip install pydantic-ai"
                ) from None

        prompt = "\n\n".join(m["content"] for m in messages if m.get("content"))
        agent = Agent(self._pydantic_model, output_type=str)
        result = agent.run_sync(prompt, model_settings=self._model_settings())
        return result.output

    def completion_structured(self, messages: list[dict[str, str]], output_type: type[T], **kwargs) -> T | None:
        """
        Call the LLM with the given messages and return a validated Pydantic model.
        Returns None if the call fails.
        """

        global Agent

        try:
            if Agent is None:
                from pydantic_ai import Agent  # type: ignore
            from pydantic_ai import UserError
        except ImportError:
            raise ImportError(
                "pydantic-ai is required for LLM support. You can install it with: pip install angr[llm] or "
                "pip install pydantic-ai"
            ) from None

        prompt = "\n\n".join(m["content"] for m in messages if m.get("content"))
        try:
            agent = Agent(self._pydantic_model, output_type=output_type)
            result = agent.run_sync(prompt, model_settings=self._model_settings())
            return result.output
        except UserError:
            l.error("Failed to get structured LLM response due to a user error.", exc_info=True)
        except Exception:  # pylint:disable=broad-exception-caught
            l.warning("Failed to get structured LLM response", exc_info=True)
        return None

    def completion_json(self, messages: list[dict[str, str]], **kwargs) -> dict | None:
        """
        Call the LLM and parse the response as JSON.
        Strips markdown code fences if present. Returns None on parse failure.
        Kept for backwards compatibility; prefer completion_structured().
        """
        text = self.completion(messages, **kwargs)
        if not text:
            return None

        # strip markdown code fences
        text = text.strip()
        text = re.sub(r"^```(?:json)?\s*\n?", "", text)
        text = re.sub(r"\n?```\s*$", "", text)
        text = text.strip()

        try:
            return json.loads(text)
        except (json.JSONDecodeError, ValueError):
            l.warning("Failed to parse LLM response as JSON: %s", text[:200])
            return None

    def __repr__(self):
        return f"<LLMClient model={self.model!r}>"



from __future__ import annotations
from . import SimLibrary
from angr.procedures import SIM_PROCEDURES as P

lib = SimLibrary()
lib.set_library_names(
    "ld.so",
    "ld-linux.so",
    "ld.so.2",
    "ld-linux.so.2",
    "ld-linux-x86-64.so.2",
    # https://git.musl-libc.org/cgit/musl/tree/configure?id=1b76ff0767d01df72f692806ee5adee13c67ef88#n322
    *[
        f"ld-musl-{arch}.so.1"
        for arch in [
            "arm",
            "aarch64",
            "nt32",
            "i386",
            "x32",
            "nt64",
            "x86_64",
            "loongarch64",
            "m68k",
            "mips64",
            "mips",
            "microblaze",
            "or1k",
            "powerpc64",
            "powerpc",
            "riscv64",
            "riscv32",
            "sh",
            "s390x",
        ]
    ],
)
lib.add_all_from_dict(P["linux_loader"])



# pylint:disable=unused-argument
from __future__ import annotations
from typing import TYPE_CHECKING
import logging

from angr.ailment import Block, AILBlockRewriter
from angr.ailment.expression import Load, Const
from angr.ailment.statement import Statement

from angr.analyses.decompiler.ailgraph_walker import AILGraphWalker
from .optimization_pass import OptimizationPass, OptimizationPassStage

if TYPE_CHECKING:
    from angr import Project


_l = logging.getLogger(name=__name__)


class BlockWalker(AILBlockRewriter):
    """
    AIL Block walker in order to perform const deref substitution
    """

    def __init__(self, project: Project):
        super().__init__()
        self._project = project
        self._new_block: Block | None = None  # output

    def _addr_belongs_to_ro_region(self, addr: int) -> bool:
        section = self._project.loader.find_section_containing(addr)
        if section is not None:
            return not section.is_writable
        segment = self._project.loader.find_segment_containing(addr)
        if segment is not None:
            return not segment.is_writable
        return False

    def _addr_belongs_to_got(self, addr: int) -> bool:
        section = self._project.loader.find_section_containing(addr)
        if section is not None:
            return section.name and "got" in section.name
        return False

    def _addr_belongs_to_object(self, addr: int) -> bool:
        obj = self._project.loader.find_object_containing(addr)
        return obj is not None

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block):
        if isinstance(expr.addr, Const):
            assert isinstance(expr.addr.value, int)
            # *(const_addr)
            # does it belong to a read-only section/segment?
            is_got = self._addr_belongs_to_got(expr.addr.value)
            if is_got or self._addr_belongs_to_ro_region(expr.addr.value):
                try:
                    w = self._project.loader.memory.unpack_word(
                        expr.addr.value,
                        expr.addr.bits // self._project.arch.byte_width,
                        endness=self._project.arch.memory_endness,
                    )
                except KeyError:
                    # we don't have enough bytes to read out
                    w = None
                if w is not None and not (is_got and w == 0):
                    # nice! replace it with the actual value
                    return Const(None, None, w, expr.bits, **expr.tags)
        elif (
            isinstance(expr.addr, Load)
            and expr.addr.bits == self._project.arch.bits
            and isinstance(expr.addr.addr, Const)
        ):
            assert isinstance(expr.addr.addr.value, int)
            # *(*(const_addr))
            # does it belong to a read-only section/segment?
            if self._addr_belongs_to_got(expr.addr.addr.value) or self._addr_belongs_to_ro_region(expr.addr.addr.value):
                w = self._project.loader.memory.unpack_word(
                    expr.addr.addr.value,
                    expr.addr.addr.bits // self._project.arch.byte_width,
                    endness=self._project.arch.memory_endness,
                )
                if w is not None and self._addr_belongs_to_object(w):
                    # nice! replace it with a load from that address
                    return Load(
                        expr.idx,
                        Const(None, None, w, expr.addr.size, **expr.addr.addr.tags),
                        expr.size,
                        expr.endness,
                        variable=expr.variable,
                        variable_offset=expr.variable_offset,
                        guard=expr.guard,
                        alt=expr.alt,
                        **expr.tags,
                    )

        return super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)


class ConstantDereferencesSimplifier(OptimizationPass):
    """
    Makes the following simplifications::

        *(*(const_addr))  ==>  *(value) iff  *const_addr == value
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_SINGLE_BLOCK_SIMPLIFICATION
    NAME = "Simplify constant dereferences"
    DESCRIPTION = __doc__.strip()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._block_walker = BlockWalker(self.project)
        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):
        # walk the entire graph and traverse each expression
        walker = AILGraphWalker(self._graph, handler=self._walk_block, replace_nodes=True)
        walker.walk()

    def _walk_block(self, block: Block) -> Block | None:
        return self._block_walker.walk(block)



from __future__ import annotations
from itertools import chain
from typing import TYPE_CHECKING, cast
from dataclasses import dataclass
from collections import defaultdict
from collections.abc import Iterable, Callable

from angr.ailment.statement import SideEffectStatement, Store, ConditionalJump, CAS
from angr.ailment.expression import (
    Const,
    Convert,
    Extract,
    Insert,
    Register,
    StackBaseOffset,
    ITE,
    VEXCCallExpression,
    Tmp,
    DirtyExpression,
    Load,
    VirtualVariable,
    Expression,
)

from angr.code_location import AILCodeLocation
from angr.engines.light import SimEngineLightAIL
from angr.knowledge_plugins.functions.function import Function
from angr.project import Project
from angr.sim_type import PointerDisposition, SimTypePointer
from angr.utils.ssa import get_reg_offset_base_and_size
from angr.calling_conventions import default_cc
from .traversal_state import TraversalState, Value
from .consts import MAX_STACK_VAR_SIZE

if TYPE_CHECKING:
    from angr.analyses.decompiler.ssailification.ssailification import Def, Kind

CUTOFF = 15  # arbitrary; be mindful of performance as various parts will be O(N^2)


def offset_sort_key(v: tuple[int | None, int]) -> tuple[int, int, int, int]:
    # used for iterating over Value in a determinisitic order
    # moderately arbitrary. feel free to change as long as
    # - it produces a total ordering, i.e. f(a) == f(b) iff a == b
    # - it never returns None values, as these cannot be sorted
    return (v[1], 0 if v[0] is not None else 1, abs(v[0] or 0), 0 if v[0] is None or v[0] < 0 else 1)


@dataclass
class DefInfo:
    """
    Information about a def, or an expression that defines a new SSA variable.
    """

    def_: Def
    kind: Kind
    loc: AILCodeLocation
    variable_offset: int  # offset from beginning of region to beginning of variable
    variable_size: int
    store_offset: int  # offset from beginning of region to beginning of store
    store_size: int
    supercedes: set[Def]

    @property
    def variable_range(self):
        return range(self.variable_offset, self.variable_offset + self.variable_size)

    @property
    def variable_endoffset(self):
        return self.variable_offset + self.variable_size


class SimEngineSSATraversal(SimEngineLightAIL[TraversalState, Value, None, None]):
    """
    This engine collects all register and stack variable locations and links them to the block of their creation.
    """

    def __init__(
        self,
        project: Project,
        simos,
        sp_tracker=None,
        bp_as_gpr: bool = False,
        stackvars: bool = False,
        use_tmps: bool = False,
        functions: Callable[[int | str], Function | None] | None = None,
    ):
        super().__init__(project)
        self.simos = simos
        self.sp_tracker = sp_tracker
        self.bp_as_gpr = bp_as_gpr
        self.stackvars = stackvars
        self.use_tmps = use_tmps
        self.functions = functions
        self.def_info: dict[Def, DefInfo] = {}
        self.pending_ptr_defines_nonlocal: dict[
            int, tuple[AILCodeLocation, StackBaseOffset, set[tuple[int, int]], bool]
        ] = {}
        # the state which skipped the last few statements; the state that breaks the loop
        self.hclb_side_exit_state: TraversalState | None = None

    def _is_top(self, expr):
        return not expr

    def _top(self, bits):
        return set()

    def _process_block_end(self, block, stmt_data, whitelist):
        # see comment in StackBaseOffset handler
        for k, lst in self.state.pending_ptr_defines.items():
            v1, v2 = lst[0]
            if k not in self.pending_ptr_defines_nonlocal:
                self.pending_ptr_defines_nonlocal[k] = (v1, v2, set(), k not in self.state.live_stackvars)
                self.state.pending_ptr_defines_nonlocal_live.add(k)
        self.state.pending_ptr_defines.clear()  # just in case

    def finalize(self):
        for stack_offset, (loc, def_, suggestions, required) in self.pending_ptr_defines_nonlocal.items():
            if not required:
                continue
            full_offset, full_endoffset = stack_offset, stack_offset + 1
            for suggested_offset, suggested_size in suggestions:
                suggested_endoffset = suggested_offset + suggested_size
                full_offset = min(full_offset, suggested_offset)
                full_endoffset = max(full_endoffset, suggested_endoffset)

            self.perform_def(
                "stack",
                def_,
                full_offset,
                full_endoffset - full_offset,
                stack_offset,
                0,
                loc,
            )

        # sketchy situation. we want to make sure everything in the extern set
        # has exactly the same extents if they overlap at all. this is sorta a redo
        # of the unify code.
        reg_mapping: defaultdict[int, set[Def]] = defaultdict(set)
        stack_mapping: defaultdict[int, set[Def]] = defaultdict(set)
        for def_, definfo in self.def_info.items():
            if not definfo.loc.is_extern:
                continue
            match definfo.kind:
                case "stack":
                    mapping = stack_mapping
                case "reg":
                    mapping = reg_mapping
                case _:
                    raise ValueError(f"Unsupported def kind {definfo.kind} in finalize")
            for suboffset in definfo.variable_range:
                mapping[suboffset].add(def_)

        current_extent = (0, 0)
        current_defs = set()

        def flush():
            for def2 in current_defs:
                definfo = self.def_info[def2]
                size2 = current_extent[1] - current_extent[0]
                definfo.variable_offset = current_extent[0]
                definfo.variable_size = size2

        for mapping in [reg_mapping, stack_mapping]:
            for offset in sorted(mapping):
                for def_ in mapping[offset]:
                    definfo = self.def_info[def_]

                    if definfo.variable_offset >= current_extent[1] or current_extent[0] == current_extent[1]:
                        flush()
                        current_defs.clear()
                        current_extent = definfo.variable_offset, definfo.variable_endoffset

                    current_defs.add(def_)
                    current_extent = (
                        min(current_extent[0], definfo.variable_offset),
                        max(current_extent[1], definfo.variable_endoffset),
                    )
        flush()

    def _acodeloc(self):
        return AILCodeLocation(
            self.block.addr,
            self.block.idx,
            self.stmt_idx,
            self.ins_addr,
        )

    def perform_def(
        self,
        kind: Kind,
        def_: Def,
        variable_offset: int,
        variable_size: int,
        store_offset: int,
        store_size: int,
        loc: AILCodeLocation | None = None,
        supercedes: Iterable[Def] | None = None,
    ):
        loc = loc or self._acodeloc()
        assert (
            loc.is_extern
            or def_ not in self.def_info
            or self.def_info[def_].loc == loc
            or self.def_info[def_].loc.is_extern
        ), "claiming an expression defines at two different locs"
        if (definfo := self.def_info.get(def_)) is None or definfo.loc.is_extern:
            definfo = DefInfo(
                def_,
                kind,
                loc,
                variable_offset,
                variable_size,
                store_offset,
                store_size,
                set(supercedes or ()),
            )
            self.def_info[def_] = definfo
        else:
            end_offset = definfo.variable_endoffset
            definfo.variable_offset = min(definfo.variable_offset, variable_offset)
            new_end_offset = max(end_offset, variable_offset + variable_size)
            definfo.variable_size = new_end_offset - definfo.variable_offset

        return definfo

    def stackvar_get(self, base_offset: int, extra_offset: int, base_size: int) -> Value:
        if extra_offset > 1 << (self.project.arch.bits - 1):
            extra_offset -= 1 << self.project.arch.bits
        concrete_offset = base_offset + extra_offset
        offset = min(concrete_offset, base_offset)
        end_offset = max(concrete_offset, base_offset) + base_size
        size = end_offset - offset
        if size >= MAX_STACK_VAR_SIZE:
            return set()

        full_offset, full_size, popped = self.state.stackvar_unify(offset, size)

        if base_offset in self.state.pending_ptr_defines_nonlocal_live:
            self.pending_ptr_defines_nonlocal.pop(base_offset, None)

        lst = self.state.pending_ptr_defines.pop(base_offset, [])
        pending_def = cast("Def", lst[-1][1]) if lst else None

        secret_stash: defaultdict[int, set[Def]] = defaultdict(set)
        while True:  # this loop should run until the UH OH is never reached
            for popped_offset in popped:
                secret_stash[popped_offset].update(self.state.stackvar_defs.pop(popped_offset, set()))
                for def2 in secret_stash[popped_offset]:
                    self.state.stackvar_defs[full_offset].add(def2)
                    definfo = self.def_info[def2]
                    if definfo.variable_offset < full_offset or definfo.variable_endoffset > full_offset + full_size:
                        # UH OH. We have information from a parallel timeline about how big this var actually is...
                        newish_offset = min(definfo.variable_offset, full_offset)
                        newish_endoffset = max(definfo.variable_endoffset, full_offset + full_size)
                        full_offset, full_size, popped2 = self.state.stackvar_unify(
                            newish_offset, newish_endoffset - newish_offset
                        )
                        popped.update(popped2)
                        break
                    definfo.variable_offset = full_offset
                    definfo.variable_size = full_size
                else:
                    continue
                break
            else:
                break

        defs: set[Def] = set().union(*secret_stash.values())
        def_as = None
        if not defs:
            if pending_def is not None:
                # if this assert trips maybe consider gating all the redef stuff on the cond?
                # maybe also consider making def nullable in case of externs
                self.perform_def(
                    "stack",
                    pending_def,
                    full_offset,
                    full_size,
                    base_offset + extra_offset,
                    base_size,
                    AILCodeLocation.make_extern(0),
                )
                def_as = {pending_def}
        else:
            def_as = defs

        if def_as is not None:
            for suboff in range(full_offset, full_offset + full_size):
                self.state.stackvar_defs[suboff] = def_as

        return self.state.live_stackvars[offset]

    def stackvar_set(self, base_offset: int, extra_offset: int, base_size: int, value: Value):
        if extra_offset > 1 << (self.project.arch.bits - 1):
            extra_offset -= 1 << self.project.arch.bits
        concrete_offset = base_offset + extra_offset
        offset = min(concrete_offset, base_offset)
        end_offset = max(concrete_offset, base_offset) + base_size
        size = end_offset - offset

        if size >= MAX_STACK_VAR_SIZE:
            return

        self.state.pending_ptr_defines_nonlocal_live.discard(base_offset)
        if base_offset in self.pending_ptr_defines_nonlocal:
            self.pending_ptr_defines_nonlocal[base_offset][2].add((offset, size))

        self.state.live_stackvars[offset] = value

        other_defs: set[Def] = set()
        liveish_defs: set[Def] = set()
        secret_stash: defaultdict[int, set[Def]] = defaultdict(set)
        reached_fixedpoint = False
        while not reached_fixedpoint:
            reached_fixedpoint = True
            for suboff in range(offset, end_offset):
                secret_stash[suboff].update(self.state.stackvar_defs.pop(suboff, set()))
                old_defs = secret_stash[suboff]
                # set up current var mapping
                self.state.stackvar_bases[suboff] = (offset, size)
                other_defs.update(old_defs)
                # additional consideration: if this def only provides values for some bytes,
                # make sure the implicated but not overwritten defs are unified
                if not base_offset + extra_offset <= suboff < base_offset + extra_offset + base_size:
                    liveish_defs.update(old_defs)
                    for old_def in old_defs:
                        definfo = self.def_info[old_def]
                        old_offset = definfo.variable_offset
                        old_end_offset = definfo.variable_endoffset

                        new_offset = min(old_offset, offset)
                        new_end_offset = max(old_end_offset, end_offset)

                        definfo.variable_offset = new_offset
                        definfo.variable_size = new_end_offset - new_offset

                        if offset != new_offset:
                            offset = new_offset
                            reached_fixedpoint = False
                        if end_offset != new_end_offset:
                            end_offset = new_end_offset
                            reached_fixedpoint = False
                        size = end_offset - offset

        loc2, def2 = self.state.pending_ptr_defines.pop(base_offset, [(None, None)])[-1]
        def_as = None
        if loc2 is not None:
            assert def2 is not None
            self.perform_def("stack", def2, offset, size, base_offset + extra_offset, base_size, loc2, other_defs)
            def_as = {def2} | liveish_defs

        if def_as is not None:
            for suboff in range(offset, end_offset):
                self.state.stackvar_defs[suboff] = def_as

    def register_get(self, offset: int, size: int, def_: Def) -> Value:
        full_offset, full_size, popped = self.state.register_unify(offset, size)

        secret_stash: defaultdict[int, set[Def]] = defaultdict(set)
        while True:  # this loop should run until the UH OH is never reached
            for popped_offset in popped:
                secret_stash[popped_offset].update(self.state.register_defs.pop(popped_offset, set()))
                for def2 in secret_stash[popped_offset]:
                    self.state.register_defs[full_offset].add(def2)
                    definfo = self.def_info[def2]
                    if definfo.variable_offset < full_offset or definfo.variable_endoffset > full_offset + full_size:
                        # UH OH. We have information from a parallel timeline about how big this var actually is...
                        newish_offset = min(definfo.variable_offset, full_offset)
                        newish_endoffset = max(definfo.variable_endoffset, full_offset + full_size)
                        full_offset, full_size, popped2 = self.state.register_unify(
                            newish_offset, newish_endoffset - newish_offset
                        )
                        popped.update(popped2)
                        break
                    definfo.variable_offset = full_offset
                    definfo.variable_size = full_size
                else:
                    continue
                break
            else:
                break

        defs: set[Def] = set().union(*secret_stash.values())
        def_as = None
        if not defs or full_offset in self.state.register_blackout:
            self.perform_def(
                "reg",
                def_,
                full_offset,
                full_size,
                offset,
                size,
                AILCodeLocation.make_extern(0) if full_offset not in self.state.register_blackout else None,
            )
            def_as = {def_}
            self.state.register_blackout.discard(full_offset)
        else:
            def_as = defs

        if def_as is not None:
            for suboff in range(full_offset, full_offset + full_size):
                self.state.register_defs[suboff] = def_as

        return self.state.live_registers[offset]

    def register_set(self, offset: int, size: int, value: Value, def_: Def):
        self.state.live_registers[offset] = value

        for suboff in range(offset, offset + size):
            self.state.register_defs.pop(suboff, None)
            self.state.register_bases[suboff] = (offset, size)

        self.perform_def("reg", def_, offset, size, offset, size)
        self.state.register_blackout.discard(offset)

        for suboff in range(offset, offset + size):
            self.state.register_defs[suboff] = {def_}

    def _handle_stmt_Assignment(self, stmt):
        src = self._expr(stmt.src)

        if isinstance(stmt.dst, Register):
            self.register_set(stmt.dst.reg_offset, stmt.dst.size, src, stmt.dst)
        elif isinstance(stmt.dst, VirtualVariable):
            self.state.live_vvars[stmt.dst.varid] = src
        elif isinstance(stmt.dst, Tmp):
            self.state.live_tmps[stmt.dst.tmp_idx] = src

    def _handle_stmt_WeakAssignment(self, stmt):
        self._expr(stmt.src)
        self._expr(stmt.dst)

    def _handle_stmt_CAS(self, stmt: CAS):
        self._expr(stmt.addr)
        self._expr(stmt.data_lo)
        if stmt.data_hi is not None:
            self._expr(stmt.data_hi)
        self._expr(stmt.expd_lo)
        if stmt.expd_hi is not None:
            self._expr(stmt.expd_hi)

    def _handle_stmt_Store(self, stmt: Store):
        src = self._expr(stmt.data)
        if stmt.guard is not None:
            self._expr(stmt.guard)

        addr = self._expr(stmt.addr)

        if self.stackvars:
            for stackoff, extra in addr:
                if stackoff is not None:
                    self.stackvar_set(stackoff, extra, stmt.data.size, src)

    def _handle_stmt_ConditionalJump(self, stmt: ConditionalJump):
        self._expr(stmt.condition)
        if stmt.true_target is not None:
            self._expr(stmt.true_target)
        if stmt.false_target is not None:
            self._expr(stmt.false_target)

        if (
            isinstance(stmt.true_target, Const)
            and isinstance(stmt.false_target, Const)
            and self.stmt_idx != len(self.block.statements) - 1
            and self.ins_addr in (stmt.true_target.value, stmt.false_target.value)
        ):
            self.hclb_side_exit_state = self.state.copy()

    def _handle_stmt_SideEffectStatement(self, stmt: SideEffectStatement):
        result = self._handle_expr_Call(stmt.expr)

        if stmt.ret_expr is not None and isinstance(stmt.ret_expr, Register):
            self.register_set(stmt.ret_expr.reg_offset, stmt.ret_expr.size, result, stmt.ret_expr)
        if stmt.fp_ret_expr is not None and isinstance(stmt.fp_ret_expr, Register):
            self.register_set(stmt.fp_ret_expr.reg_offset, stmt.fp_ret_expr.size, result, stmt.fp_ret_expr)

    def _handle_expr_Call(self, expr):
        target = expr.target
        if isinstance(target, Expression):
            self._expr(target)

        def_size = None
        def_size_arg = None
        if target == "memset":
            # TODO other intrinsics that do a def
            assert expr.args is not None
            def_size_values = self._expr(expr.args[2])
            if len(def_size_values) == 1 and max(def_size_values)[0] is None:
                def_size = max(def_size_values)[1]
                def_size_arg = expr.args[2]

        if isinstance(target, Const) and isinstance(target.value, int):
            target = target.value
        target = self.functions(target) if self.functions is not None and isinstance(target, (str, int)) else None
        if expr.prototype is not None:
            proto = expr.prototype
        elif target is not None and target.prototype is not None:
            proto = target.prototype
        else:
            proto = None

        if proto is not None:
            extra_nones = [None] * max(len(expr.args or []) - len(proto.args), 0)
            for ty, argexpr in zip(chain(proto.args, extra_nones), expr.args or []):
                if argexpr is def_size_arg:
                    assert def_size is not None
                    value = {(None, def_size)}
                else:
                    value = self._expr(argexpr)

                if not value:
                    continue

                if not isinstance(ty, SimTypePointer):
                    continue
                if ty.disposition == PointerDisposition.OUT:
                    outptr = True
                elif ty.disposition in (PointerDisposition.NONE, PointerDisposition.UNKNOWN):
                    continue
                else:
                    outptr = False

                if outptr and def_size is not None:
                    size = def_size
                else:
                    if ty.pts_to is None:
                        size = 1
                    else:
                        size = (ty.pts_to.size or 8) // 8
                        if not size:
                            continue

                for stackref, extra in value:
                    if stackref is not None:
                        if outptr:
                            self.stackvar_set(stackref, extra, size, set())
                        else:
                            self.stackvar_get(stackref, extra, size)
        else:
            for argexpr in expr.args or []:
                self._expr(argexpr)

        # kill caller-saved registers
        if expr.calling_convention is not None:
            cc = expr.calling_convention
        elif target is not None and target.calling_convention is not None:
            cc = target.calling_convention
        else:
            cc = default_cc(self.arch.name, platform=self.simos.name if self.simos is not None else None)
            assert cc is not None
            cc = cc(self.arch)

        for reg_name in cc.CALLER_SAVED_REGS:
            reg_offset, _ = self.arch.registers[reg_name]
            base_off, base_size = get_reg_offset_base_and_size(reg_offset, self.arch)
            self.state.live_registers.pop(base_off, None)
            self.state.register_blackout.add(base_off)
            for suboff in range(base_off, base_off + base_size):
                self.state.register_defs.pop(suboff, None)
        for reg in cc.arch.vex_cc_regs or []:
            self.state.live_registers.pop(reg.vex_offset, None)
            self.state.register_blackout.add(reg.vex_offset)
            for suboff in range(reg.vex_offset, reg.vex_offset + reg.size):
                self.state.register_defs.pop(suboff, None)

        return set()

    def _handle_stmt_Dummy(self, stmt):
        pass

    def _handle_stmt_DirtyStatement(self, stmt):
        self._expr(stmt.dirty)

    def _handle_stmt_Jump(self, stmt):
        self._expr(stmt.target)

    _handle_stmt_Label = _handle_stmt_Dummy

    def _handle_stmt_Return(self, stmt):
        for expr in stmt.ret_exprs:
            self._expr(expr)

    def _handle_expr_VirtualVariable(self, expr):
        return self.state.live_vvars[expr.varid]

    def _handle_expr_Register(self, expr: Register):
        return self.register_get(expr.reg_offset, expr.size, expr)

    def _handle_expr_Load(self, expr: Load):
        addr = self._expr(expr.addr)
        result = set()
        if self.stackvars:
            for offset, extra in addr:
                if offset is None:
                    continue
                result.update(self.stackvar_get(offset, extra, expr.size))
        return result

    def _handle_expr_StackBaseOffset(self, expr: StackBaseOffset) -> Value:
        if not self.stackvars:
            return set()
        if not isinstance(expr.offset, int):
            return set()

        if expr.offset not in self.state.pending_ptr_defines:
            self.state.pending_ptr_defines[expr.offset] = [(self._acodeloc(), expr)]
            # if this is the first time we've ever seen anything resembling this stackvar,
            # force this to be a def. otherwise, track future uses of this reference.
            # if we get to the end of the block and we haven't seen it used yet,
            # we'll require no more reads-before-writes in the rest of the function
            # to call this a new def
            if expr.offset not in self.state.stackvar_bases:
                self.stackvar_get(expr.offset, 0, 1)
                self.state.pending_ptr_defines[expr.offset] = [(self._acodeloc(), expr)]
        else:
            self.state.pending_ptr_defines[expr.offset].append((self._acodeloc(), expr))

        return {(expr.offset, 0)}

    def _handle_expr_Tmp(self, expr: Tmp):
        return self.state.live_tmps[expr.tmp_idx]

    def _handle_expr_Const(self, expr) -> Value:
        if isinstance(expr.value, int):
            return {(None, expr.value)}
        return set()

    def _handle_expr_Convert(self, expr: Convert) -> Value:
        val = self._expr(expr.operand)
        result = set()
        for off, v in val:
            if off is not None:
                continue
            if expr.is_signed and v > 1 << (expr.from_bits - 1):
                v -= 1 << expr.from_bits
            result.add((None, v))
        return result

    def _handle_expr_Reinterpret(self, expr) -> Value:
        self._expr(expr.operand)
        return set()

    def _handle_expr_UnaryOp(self, expr) -> Value:
        self._expr(expr.operand)
        return set()

    def _handle_expr_BinaryOp(self, expr) -> Value:
        a0 = self._expr(expr.operands[0])
        a1 = self._expr(expr.operands[1])

        if expr.op == "Add":
            sign = 1
        elif expr.op == "Sub":
            sign = -1
        else:
            return set()

        result: Value = set()
        for arg0 in a0:
            for arg1 in a1:
                if (arg0[0] is None or arg1[0] is None) and (arg0[1] is not None and arg1[1] is not None):
                    result.add((arg0[0] or arg1[0], arg0[1] + sign * arg1[1]))
        if len(result) > CUTOFF:
            result = set(sorted(result, key=offset_sort_key)[:CUTOFF])
        return result

    def _handle_expr_Phi(self, expr) -> Value:
        result = set()
        for _, vvar in expr.src_and_vvars:
            if vvar is not None:
                result.update(self._expr(vvar))
        if len(result) > CUTOFF:
            result = set(sorted(result, key=offset_sort_key)[:CUTOFF])
        return result

    def _handle_expr_ITE(self, expr: ITE):
        self._expr(expr.cond)
        self._expr(expr.iftrue)
        self._expr(expr.iffalse)
        return set()

    def _handle_expr_VEXCCallExpression(self, expr: VEXCCallExpression):
        for operand in expr.operands:
            self._expr(operand)
        return set()

    def _handle_expr_DirtyExpression(self, expr: DirtyExpression):
        for operand in expr.operands:
            self._expr(operand)
        if expr.guard is not None:
            self._expr(expr.guard)
        if expr.maddr is not None:
            self._expr(expr.maddr)
        return set()

    def _handle_expr_Extract(self, expr: Extract):
        self._expr(expr.base)
        self._expr(expr.offset)
        return set()

    def _handle_expr_Insert(self, expr: Insert):
        self._expr(expr.base)
        self._expr(expr.offset)
        self._expr(expr.value)
        return set()

    # pylint: disable=unused-argument,no-self-use
    def _handle_expr_MultiStatementExpression(self, expr) -> Value:
        return set()

    def _handle_expr_BasePointerOffset(self, expr) -> Value:
        return set()

    def _unreachable(self, *args, **kwargs):
        assert False, "unreachable"

    _handle_binop_CmpLE = _unreachable
    _handle_binop_CmpLT = _unreachable
    _handle_binop_CmpGE = _unreachable
    _handle_binop_CmpGT = _unreachable
    _handle_binop_CmpEQ = _unreachable
    _handle_binop_CmpNE = _unreachable
    _handle_binop_CmpORD = _unreachable
    _handle_binop_Add = _unreachable
    _handle_binop_AddF = _unreachable
    _handle_binop_AddV = _unreachable
    _handle_binop_And = _unreachable
    _handle_binop_Carry = _unreachable
    _handle_binop_CmpF = _unreachable
    _handle_binop_Concat = _unreachable
    _handle_binop_Div = _unreachable
    _handle_binop_DivF = _unreachable
    _handle_binop_DivV = _unreachable
    _handle_binop_LogicalAnd = _unreachable
    _handle_binop_LogicalOr = _unreachable
    _handle_binop_Mod = _unreachable
    _handle_binop_Mul = _unreachable
    _handle_binop_Mull = _unreachable
    _handle_binop_MulF = _unreachable
    _handle_binop_MulV = _unreachable
    _handle_binop_MulHiV = _unreachable
    _handle_binop_Or = _unreachable
    _handle_binop_Rol = _unreachable
    _handle_binop_Ror = _unreachable
    _handle_binop_SBorrow = _unreachable
    _handle_binop_SCarry = _unreachable
    _handle_binop_Sar = _unreachable
    _handle_binop_Shl = _unreachable
    _handle_binop_Shr = _unreachable
    _handle_binop_Sub = _unreachable
    _handle_binop_SubF = _unreachable
    _handle_binop_SubV = _unreachable
    _handle_binop_Xor = _unreachable
    _handle_binop_InterleaveLOV = _unreachable
    _handle_binop_InterleaveHIV = _unreachable
    _handle_binop_CasCmpEQ = _unreachable
    _handle_binop_CasCmpNE = _unreachable
    _handle_binop_ExpCmpNE = _unreachable
    _handle_binop_SarNV = _unreachable
    _handle_binop_ShrNV = _unreachable
    _handle_binop_ShlNV = _unreachable
    _handle_binop_CmpEQV = _unreachable
    _handle_binop_CmpNEV = _unreachable
    _handle_binop_CmpGEV = _unreachable
    _handle_binop_CmpGTV = _unreachable
    _handle_binop_CmpLEV = _unreachable
    _handle_binop_CmpLTV = _unreachable
    _handle_binop_MinV = _unreachable
    _handle_binop_MaxV = _unreachable
    _handle_binop_QAddV = _unreachable
    _handle_binop_QNarrowBinV = _unreachable
    _handle_binop_PermV = _unreachable
    _handle_binop_Set = _unreachable

    _handle_unop_BitwiseNeg = _unreachable
    _handle_unop_Dereference = _unreachable
    _handle_unop_Neg = _unreachable
    _handle_unop_Not = _unreachable
    _handle_unop_Reference = _unreachable
    _handle_unop_Clz = _unreachable
    _handle_unop_Ctz = _unreachable
    _handle_unop_GetMSBs = _unreachable
    _handle_unop_unpack = _unreachable
    _handle_unop_Sqrt = _unreachable
    _handle_unop_RSqrtEst = _unreachable



from __future__ import annotations
import angr


class stpcpy(angr.SimProcedure):
    """stpcpy"""

    # pylint:disable=arguments-differ

    def run(self, dst, src):
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]
        strncpy = angr.SIM_PROCEDURES["libc"]["strncpy"]
        src_len = self.inline_call(strlen, src)

        return (
            self.inline_call(strncpy, dst, src, src_len.ret_expr + 1, src_len=src_len.ret_expr).ret_expr
            + src_len.ret_expr
        )



from __future__ import annotations
from itertools import count

import claripy

import angr

from angr import sim_options as o
from angr.state_plugins.sim_action import SimActionData

fastpath_data_counter = count()


class receive(angr.SimProcedure):
    # pylint:disable=arguments-differ,attribute-defined-outside-init,missing-class-docstring,redefined-outer-name

    def run(self, fd, buf, count, rx_bytes):
        if o.CGC_ENFORCE_FD in self.state.options:
            fd = 0

        simfd = self.state.posix.get_fd(fd)
        if simfd is None:
            return -1

        if self.state.mode == "fastpath":
            # Special case for CFG generation
            if not self.state.solver.symbolic(count):
                data = self.state.solver.Unconstrained(
                    f"receive_data_{next(fastpath_data_counter)}", self.state.solver.eval_one(count) * 8
                )
                self.state.memory.store(buf, data)
            else:
                count = self.state.solver.Unconstrained("receive_length", self.state.arch.bits)
            self.state.memory.store(rx_bytes, count, endness="Iend_LE")

            return claripy.BVV(0, self.state.arch.bits)

        # check invalid memory accesses
        # rules for invalid: greater than 0xc0 or wraps around
        if self.state.solver.max_int(buf + count) > 0xC0000000 or self.state.solver.min_int(
            buf + count
        ) < self.state.solver.min_int(buf):
            return 2
        try:
            writable = self.state.solver.eval(self.state.memory.permissions(self.state.solver.eval(buf))) & 2 != 0
        except angr.SimMemoryError:
            writable = False
        if not writable:
            return 2

        if o.CGC_NO_SYMBOLIC_RECEIVE_LENGTH in self.state.options:
            count = self.state.solver.eval(count)
            if self.state.cgc.max_receive_size > 0:
                count = min(count, self.state.cgc.max_receive_size)

            do_concrete_update = (
                o.UNICORN_HANDLE_SYMBOLIC_ADDRESSES in self.state.options
                or o.UNICORN_HANDLE_SYMBOLIC_CONDITIONS in self.state.options
            )
            read_length = simfd.read(buf, count, short_reads=False, do_concrete_update=do_concrete_update)
            if type(read_length) is int:
                read_length = claripy.BVV(read_length, 32)
            self.state.memory.store(rx_bytes, read_length, condition=rx_bytes != 0, endness="Iend_LE")
            self.size = read_length

            return 0
        if self.state.solver.solution(count != 0, True):
            data, read_length = simfd.read_data(count)
            if not self.state.solver.is_true(read_length == 0):
                self.state.memory.store(buf, data, size=read_length)
            action_list = list(self.state.history.recent_actions)

            try:
                # get and fix up the memory write
                action = next(
                    a
                    for a in reversed(action_list)
                    if isinstance(a, SimActionData) and a.action == "write" and a.type == "mem"
                )
                action.size.ast = read_length
                action.data.ast = action.actual_value.ast
                self.data = data
            except StopIteration:
                # the write didn't occur (i.e., size of 0)
                self.data = None
        else:
            self.data = None

        self.size = read_length
        if type(read_length) is int:
            read_length = claripy.BVV(read_length, 32)
        self.state.memory.store(rx_bytes, read_length, condition=rx_bytes != 0, endness="Iend_LE")
        return 0



from __future__ import annotations
from itertools import chain
from typing import TYPE_CHECKING
from collections.abc import Iterable
import logging

import pyvex
import claripy

from angr.storage.memory_mixins.paged_memory.pages.multi_values import MultiValues, mv_is_bv
from angr.engines.light import SimEngineNostmtVEX, SpOffset
from angr.engines.vex.claripy.datalayer import value as claripy_value
from angr.errors import SimEngineError, SimMemoryMissingError
from angr.utils.constants import DEFAULT_STATEMENT
from angr.knowledge_plugins.key_definitions.definition import Definition
from angr.knowledge_plugins.key_definitions.live_definitions import LiveDefinitions
from angr.knowledge_plugins.key_definitions.tag import LocalVariableTag, ParameterTag, Tag
from angr.knowledge_plugins.key_definitions.atoms import Atom, Register, MemoryLocation, Tmp
from angr.knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
from angr.knowledge_plugins.key_definitions.heap_address import HeapAddress
from angr.code_location import CodeLocation, ExternalCodeLocation
from .rd_state import ReachingDefinitionsState
from .function_handler import FunctionCallData

if TYPE_CHECKING:
    from angr.knowledge_plugins import FunctionManager
    from .function_handler import FunctionHandler


l = logging.getLogger(name=__name__)

unop_handler = SimEngineNostmtVEX[
    ReachingDefinitionsState, MultiValues[claripy.ast.BV | claripy.ast.FP], ReachingDefinitionsState
].unop_handler
binop_handler = SimEngineNostmtVEX[
    ReachingDefinitionsState, MultiValues[claripy.ast.BV | claripy.ast.FP], ReachingDefinitionsState
].binop_handler


class SimEngineRDVEX(
    SimEngineNostmtVEX[
        ReachingDefinitionsState, MultiValues[claripy.ast.BV | claripy.ast.FP], ReachingDefinitionsState
    ],
):  # pylint:disable=abstract-method
    """
    Implements the VEX execution engine for reaching definition analysis.
    """

    def __init__(self, project, function_handler: FunctionHandler, functions: FunctionManager):
        super().__init__(project)
        self.functions = functions
        self._function_handler = function_handler
        self._visited_blocks = None
        self._dep_graph = None

        self.state: ReachingDefinitionsState

    def process(
        self, state, *, block=None, fail_fast=False, visited_blocks=None, dep_graph=None, whitelist=None, **kwargs
    ):
        self._visited_blocks = visited_blocks
        self._dep_graph = dep_graph
        # we are using a completely different state. Therefore, we directly call our _process() method before
        # SimEngine becomes flexible enough.
        try:
            return super().process(
                state,
                whitelist=whitelist,
                block=block,
            )
        except SimEngineError as e:
            if fail_fast is True:
                raise e
            l.error(e)
        return self.state

    def _process_block_end(self, stmt_result, whitelist):
        self.stmt_idx = DEFAULT_STATEMENT
        self._set_codeloc()

        function_handled = False
        if self.block.vex.jumpkind == "Ijk_Call" and self.block.vex.next is not None:
            # it has to be a function
            block_next = self.block.vex.next
            assert isinstance(block_next, pyvex.expr.IRExpr)
            addr = self._expr_bv(block_next)
            self._handle_function(addr)
            function_handled = True
        elif self.block.vex.jumpkind == "Ijk_Boring" and self.block.vex.next is not None:
            # test if the target addr is a function or not
            block_next = self.block.vex.next
            assert isinstance(block_next, pyvex.expr.IRExpr)
            addr = self._expr_bv(block_next)
            addr_v = addr.one_value()
            if addr_v is not None and addr_v.concrete:
                addr_int = addr_v.concrete_value
                if addr_int in self.functions:
                    # yes it's a jump to a function
                    self._handle_function(addr)
                    function_handled = True

        # take care of OP_AFTER during statement processing for function calls in a block
        if self.state.analysis and function_handled and self.block.vex.statements:
            self.state.analysis.stmt_observe(
                self.stmt_idx, self.block.vex.statements[-1], self.block, self.state, OP_AFTER
            )
            self.state.analysis.insn_observe(
                self.ins_addr, self.block.vex.statements[-1], self.block, self.state, OP_AFTER
            )

        return self.state

    #
    # Private methods
    #

    def _expr_bv(self, expr: pyvex.expr.IRExpr) -> MultiValues[claripy.ast.BV]:
        result = self._expr(expr)
        assert mv_is_bv(result)
        return result

    def _expr_pair(
        self, arg0: pyvex.expr.IRExpr, arg1: pyvex.expr.IRExpr
    ) -> (
        tuple[MultiValues[claripy.ast.BV], MultiValues[claripy.ast.BV]]
        | tuple[MultiValues[claripy.ast.FP], MultiValues[claripy.ast.FP]]
    ):
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        assert type(r0) is type(r1)
        return r0, r1  # type: ignore

    def _external_codeloc(self):
        return ExternalCodeLocation(self.state.codeloc.context)

    def _set_codeloc(self):
        # TODO do we want a better mechanism to specify context updates?
        new_codeloc = CodeLocation(
            self.block.addr, self.stmt_idx, ins_addr=self.ins_addr, context=self.state.codeloc.context
        )
        self.state.move_codelocs(new_codeloc)
        self.state.analysis.model.at_new_stmt(new_codeloc)

    def _is_top(self, expr):
        """
        MultiValues are not really "top" in the stricter sense. They are just a collection of values,
        some of which might be top
        """
        return False

    def _top(self, bits) -> MultiValues[claripy.ast.BV | claripy.ast.FP]:
        return MultiValues(self.state.top(bits))

    #
    # VEX statement handlers
    #

    def _stmt(self, stmt):
        if self.state.analysis:
            self.state.analysis.stmt_observe(self.stmt_idx, stmt, self.block, self.state, OP_BEFORE)
            self.state.analysis.insn_observe(self.ins_addr, stmt, self.block, self.state, OP_BEFORE)

        self._set_codeloc()
        result = super()._stmt(stmt)

        if self.state.analysis:
            self.state.analysis.stmt_observe(self.stmt_idx, stmt, self.block, self.state, OP_AFTER)
            self.state.analysis.insn_observe(self.ins_addr, stmt, self.block, self.state, OP_AFTER)

        return result

    def _handle_stmt_WrTmp(self, stmt):
        data = self._expr(stmt.data)

        tmp_atom = Tmp(stmt.tmp, self.tyenv.sizeof(stmt.tmp) // self.arch.byte_width)
        # if len(data.values) == 1 and 0 in data.values:
        #     data_v = data.one_value()
        #     if data_v is not None:
        #         # annotate data with its definition
        #         data = MultiValues(offset_to_values={
        #             0: {self.state.annotate_with_def(data_v, Definition(tmp_atom, self._codeloc()))
        #                 }
        #         })
        self.tmps[stmt.tmp] = data

        self.state.kill_and_add_definition(
            tmp_atom,
            data,
        )

    # e.g. PUT(rsp) = t2, t2 might include multiple values
    def _handle_stmt_Put(self, stmt):
        size: int = stmt.data.result_size(self.tyenv) // 8
        reg = Register(stmt.offset, size, self.arch)
        data = self._expr(stmt.data)

        # special handling for references to heap or stack variables
        if data.count() == 1:
            for d in next(iter(data.values())):
                if self.state.is_heap_address(d):
                    heap_offset = self.state.get_heap_offset(d)
                    if heap_offset is not None:
                        self.state.add_heap_use(heap_offset, 1)
                elif self.state.is_stack_address(d):
                    stack_offset = self.state.get_stack_offset(d)
                    if stack_offset is not None:
                        self.state.add_stack_use(stack_offset, 1)

        if self.state.exit_observed and stmt.offset == self.arch.sp_offset:
            return
        self.state.kill_and_add_definition(reg, data)

    def _handle_stmt_PutI(self, stmt):
        pass

    # e.g. STle(t6) = t21, t6 and/or t21 might include multiple values
    def _handle_stmt_Store(self, stmt):
        addr = self._expr_bv(stmt.addr)
        size = stmt.data.result_size(self.tyenv) // 8
        data = self._expr(stmt.data)

        if addr.count() == 1:
            addrs = next(iter(addr.values()))
            self._store_core(addrs, size, data, endness=stmt.endness)

    def _handle_stmt_StoreG(self, stmt):
        guard = self._expr_bv(stmt.guard)
        guard_v = guard.one_value()

        if guard_v is not None and claripy.is_true(guard_v != 0):
            addr = self._expr_bv(stmt.addr)
            if addr.count() == 1:
                addrs = next(iter(addr.values()))
                size = stmt.data.result_size(self.tyenv) // 8
                data = self._expr(stmt.data)
                self._store_core(addrs, size, data)
        elif guard_v is not None and claripy.is_false(guard_v != 0):
            pass
        else:
            # guard.data == {True, False}
            # get current data
            addr = self._expr_bv(stmt.addr)
            if addr.count() == 1:
                addrs = next(iter(addr.values()))
                size = stmt.data.result_size(self.tyenv) // 8
                data_old = self._load_core(addrs, size, stmt.endness)
                data = self._expr(stmt.data)

                self._store_core(addrs, size, data, data_old=data_old)

    def _store_core(
        self,
        addr: Iterable[int | claripy.ast.bv.BV],
        size: int,
        data: MultiValues,
        data_old: MultiValues | None = None,
        endness=None,
    ):
        if data_old is not None:
            data = data.merge(data_old)

        for a in addr:
            if self.state.is_top(a):
                l.debug("Memory address undefined, ins_addr = %#x.", self.ins_addr)
            else:
                tags: set[Tag] | None
                if isinstance(a, int):
                    atom = MemoryLocation(a, size)
                    tags = None
                elif self.state.is_stack_address(a):
                    offset = self.state.get_stack_offset(a)
                    if offset is None:
                        continue
                    atom = MemoryLocation(SpOffset(self.arch.bits, offset), size)
                    function_address = None  # we cannot get the function address in the middle of a store if a CFG
                    # does not exist. you should backpatch the function address later using
                    # the 'ins_addr' metadata entry.
                    tags = {
                        LocalVariableTag(
                            function=function_address,
                            metadata={"tagged_by": "SimEngineRDVEX._store_core", "ins_addr": self.ins_addr},
                        )
                    }

                elif self.state.is_heap_address(a):
                    offset = self.state.get_heap_offset(a)
                    if offset is not None:
                        atom = MemoryLocation(HeapAddress(offset), size)
                        tags = None
                    else:
                        continue

                elif isinstance(a, claripy.ast.BV):
                    addr_v = a.concrete_value
                    atom = MemoryLocation(addr_v, size)
                    tags = None

                else:
                    continue

                # different addresses are not killed by a subsequent iteration, because kill only removes entries
                # with same index and same size
                self.state.kill_and_add_definition(atom, data, tags=tags, endness=endness)

    def _handle_stmt_LoadG(self, stmt):
        guard = self._expr_bv(stmt.guard)
        guard_v = guard.one_value()

        if guard_v is not None and claripy.is_true(guard_v != 0):
            # FIXME: full conversion support
            if stmt.cvt.find("Ident") < 0:
                l.warning("Unsupported conversion %s in LoadG.", stmt.cvt)
            load_expr = pyvex.expr.Load(stmt.end, stmt.cvt_types[1], stmt.addr)
            wr_tmp_stmt = pyvex.stmt.WrTmp(stmt.dst, load_expr)
            self._handle_stmt_WrTmp(wr_tmp_stmt)
        elif guard_v is not None and claripy.is_false(guard_v != 0):
            wr_tmp_stmt = pyvex.stmt.WrTmp(stmt.dst, stmt.alt)
            self._handle_stmt_WrTmp(wr_tmp_stmt)
        else:
            if stmt.cvt.find("Ident") < 0:
                l.warning("Unsupported conversion %s in LoadG.", stmt.cvt)
            load_expr = pyvex.expr.Load(stmt.end, stmt.cvt_types[1], stmt.addr)

            load_expr_v = self._expr(load_expr)
            alt_v = self._expr(stmt.alt)

            data = load_expr_v.merge(alt_v)
            self.state.kill_and_add_definition(Tmp(stmt.dst, self.tyenv.sizeof(stmt.dst)), data)

    def _handle_stmt_Exit(self, stmt):
        _ = self._expr(stmt.guard)
        target = stmt.dst.value
        self.state.mark_guard(target)
        if self.state.analysis is not None:
            self.state.analysis.exit_observe(
                self.block.addr,
                self.stmt_idx,
                self.block,
                self.state,
            )
        if (
            self.block.instruction_addrs
            and self.ins_addr in self.block.instruction_addrs
            and self.block.instruction_addrs.index(self.ins_addr) == self.block.instructions - 1
        ):
            self.state.exit_observed = True

    def _handle_stmt_IMark(self, stmt):
        pass

    def _handle_stmt_AbiHint(self, stmt):
        pass

    def _handle_stmt_LLSC(self, stmt):
        if stmt.storedata is None:
            # load-link
            addr = self._expr_bv(stmt.addr)
            if addr.count() == 1:
                addrs = next(iter(addr.values()))
                size = self.tyenv.sizeof(stmt.result) // self.arch.byte_width
                load_result = self._load_core(addrs, size, stmt.endness)
                self.tmps[stmt.result] = load_result
                self.state.kill_and_add_definition(
                    Tmp(stmt.result, self.tyenv.sizeof(stmt.result) // self.arch.byte_width),
                    load_result,
                )
        else:
            # store-conditional
            storedata = self._expr(stmt.storedata)
            addr = self._expr_bv(stmt.addr)
            if addr.count() == 1:
                addrs = next(iter(addr.values()))
                if isinstance(stmt.storedata, pyvex.expr.Const):
                    size = stmt.storedata.con.size // self.arch.byte_width
                else:
                    assert isinstance(stmt.storedata, pyvex.expr.RdTmp)
                    size = self.tyenv.sizeof(stmt.storedata.tmp) // self.arch.byte_width

                self._store_core(addrs, size, storedata)
                self.tmps[stmt.result] = MultiValues(claripy.BVV(1, 1))
                self.state.kill_and_add_definition(
                    Tmp(stmt.result, self.tyenv.sizeof(stmt.result) // self.arch.byte_width),
                    self.tmps[stmt.result],
                )

    #
    # VEX expression handlers
    #

    def _handle_expr_RdTmp(self, expr: pyvex.expr.RdTmp):
        self.state.add_tmp_use(expr.tmp)

        if expr.tmp in self.tmps:
            return self.tmps[expr.tmp]
        return self._top(pyvex.get_type_size(self.tyenv.lookup(expr.tmp)))

    # e.g. t0 = GET:I64(rsp), rsp might be defined multiple times
    def _handle_expr_Get(self, expr: pyvex.expr.Get):
        bits: int = expr.result_size(self.tyenv)
        size: int = bits // self.arch.byte_width

        reg_atom = Register(expr.offset, size, self.arch)
        try:
            values: MultiValues = self.state.registers.load(expr.offset, size=size)
        except SimMemoryMissingError:
            top = self.state.top(size * self.arch.byte_width)
            # annotate it
            top = self.state.annotate_with_def(top, Definition(reg_atom, self._external_codeloc()))
            values = MultiValues(top)
            # write it to registers
            self.state.kill_and_add_definition(reg_atom, values, override_codeloc=self._external_codeloc())

        current_defs: Iterable[Definition[Atom]] | None = None
        for vs in values.values():
            for v in vs:
                if current_defs is None:
                    current_defs = self.state.extract_defs(v)
                else:
                    current_defs = chain(current_defs, self.state.extract_defs(v))

        assert current_defs is not None
        self.state.add_register_use_by_defs(current_defs)

        return values

    def _handle_expr_GetI(self, expr) -> MultiValues[claripy.ast.BV | claripy.ast.FP]:
        return MultiValues(self.state.top(expr.result_size(self.tyenv)))

    # e.g. t27 = LDle:I64(t9), t9 might include multiple values
    # caution: Is also called from StoreG
    def _handle_expr_Load(self, expr) -> MultiValues[claripy.ast.BV | claripy.ast.FP]:
        addr = self._expr_bv(expr.addr)
        bits = expr.result_size(self.tyenv)
        size = bits // self.arch.byte_width

        # convert addr from MultiValues to a list of valid addresses
        if addr.count() == 1 and 0 in addr:
            addrs = list(addr[0])
            return self._load_core(addrs, size, expr.endness)

        top = self.state.top(bits)
        # annotate it
        dummy_atom = MemoryLocation(0, size)
        def_ = Definition(dummy_atom, self._external_codeloc())
        top = self.state.annotate_with_def(top, def_)
        # add use
        self.state.add_memory_use_by_def(def_)
        return MultiValues(top)

    def _load_core(
        self, addrs: Iterable[claripy.ast.BV], size: int, endness: str
    ) -> MultiValues[claripy.ast.BV | claripy.ast.FP]:
        result: MultiValues | None = None
        # we may get more than one stack addrs with the same value but different annotations (because they are defined
        # at different locations). only load them once.
        loaded_stack_offsets = set()

        for addr in addrs:
            if self.state.is_top(addr):
                l.debug("Memory address undefined, ins_addr = %#x.", self.ins_addr)
            elif self.state.is_stack_address(addr):
                # Load data from a local variable
                stack_offset = self.state.get_stack_offset(addr)
                if stack_offset is not None and stack_offset not in loaded_stack_offsets:
                    loaded_stack_offsets.add(stack_offset)
                    stack_addr = self.state.live_definitions.stack_offset_to_stack_addr(stack_offset)
                    try:
                        vs: MultiValues = self.state.stack.load(stack_addr, size=size, endness=endness)
                        # extract definitions
                        defs = set(LiveDefinitions.extract_defs_from_mv(vs))
                    except SimMemoryMissingError:
                        continue

                    self.state.add_stack_use_by_defs(defs)
                    result = result.merge(vs) if result is not None else vs

            elif self.state.is_heap_address(addr):
                # Load data from the heap
                heap_offset = self.state.get_heap_offset(addr)
                if heap_offset is not None:
                    try:
                        vs: MultiValues = self.state.heap.load(heap_offset, size=size, endness=endness)
                        defs = set(LiveDefinitions.extract_defs_from_mv(vs))
                    except SimMemoryMissingError:
                        continue

                    self.state.add_heap_use_by_defs(defs)
                    result = result.merge(vs) if result is not None else vs

            else:
                addr_v = addr.concrete_value

                # Load data from a global region
                try:
                    vs: MultiValues = self.state.memory.load(addr_v, size=size, endness=endness)
                    defs = set(LiveDefinitions.extract_defs_from_mv(vs))
                except SimMemoryMissingError:
                    try:
                        val = self.project.loader.memory.unpack_word(addr_v, size=size)
                        section = self.project.loader.find_section_containing(addr_v)
                        missing_atom = MemoryLocation(addr_v, size)
                        missing_def = Definition(missing_atom, self._external_codeloc())
                        if val == 0 and (not section or section.is_writable):
                            top = self.state.top(size * self.arch.byte_width)
                            v = self.state.annotate_with_def(top, missing_def)
                        else:
                            v = self.state.annotate_with_def(claripy.BVV(val, size * self.arch.byte_width), missing_def)
                        vs = MultiValues(v)
                        if not section or section.is_writable:
                            self.state.memory.store(addr_v, vs, size=size, endness=endness)
                            self.state.all_definitions.add(missing_def)
                        defs = {missing_def}
                    except KeyError:
                        continue

                self.state.add_memory_use_by_defs(defs)
                result = result.merge(vs) if result is not None else vs

        if result is None:
            result = MultiValues(self.state.top(size * self.arch.byte_width))

        return result

    # CAUTION: experimental
    def _handle_expr_ITE(self, expr):
        cond = self._expr(expr.cond)
        cond_v = cond.one_value()
        iftrue = self._expr(expr.iftrue)
        iffalse = self._expr(expr.iffalse)

        if claripy.is_true(cond_v):
            return iftrue
        if claripy.is_false(cond_v):
            return iffalse
        return iftrue.merge(iffalse)

    #
    # Unary operation handlers
    #

    def _handle_expr_Const(self, expr):
        clrp = claripy_value(expr.con.type, expr.con.value)
        self.state.mark_const(expr.con.value, len(clrp) // 8)
        return MultiValues(clrp)

    def _handle_conversion(self, from_size, to_size, signed, operand) -> MultiValues[claripy.ast.BV | claripy.ast.FP]:
        arg_0 = self._expr_bv(operand)

        # if there are multiple values with only one offset, we apply conversion to each one of them
        # otherwise, we return a TOP

        if arg_0.count() == 1:
            # extension, extract, or doing nothing
            data: set[claripy.ast.BV | claripy.ast.FP] = set()
            for v in next(iter(arg_0.values())):
                assert v.size() == from_size
                if to_size > from_size:
                    if signed:
                        data.add(v.sign_extend(to_size - from_size))
                    else:
                        data.add(v.zero_extend(to_size - from_size))
                else:
                    data.add(v[to_size - 1 : 0])
            r = MultiValues({next(iter(arg_0.keys())): data})

        else:
            r = self._top(to_size)

        return r

    @unop_handler
    def _handle_unop_Not(self, expr: pyvex.expr.Unop) -> MultiValues:
        arg0 = expr.args[0]
        expr_0 = self._expr_bv(arg0)
        bits = expr.result_size(self.tyenv)

        e0 = expr_0.one_value()

        if e0 is not None and not e0.symbolic:
            return MultiValues(~e0)  # pylint:disable=invalid-unary-operand-type

        return MultiValues(self.state.top(bits))

    @unop_handler
    def _handle_unop_Clz(self, expr: pyvex.expr.Unop) -> MultiValues:
        arg0 = expr.args[0]
        _ = self._expr(arg0)
        bits = expr.result_size(self.tyenv)
        # Need to actually implement this later
        return MultiValues(self.state.top(bits))

    @unop_handler
    def _handle_unop_Ctz(self, expr: pyvex.expr.Unop) -> MultiValues:
        arg0 = expr.args[0]
        _ = self._expr(arg0)
        bits = expr.result_size(self.tyenv)
        # Need to actually implement this later
        return MultiValues(self.state.top(bits))

    #
    # Binary operation handlers
    #
    @binop_handler
    def _handle_binop_ExpCmpNE64(self, expr: pyvex.expr.Binop) -> MultiValues:
        _, _ = self._expr(expr.args[0]), self._expr(expr.args[1])
        bits = expr.result_size(self.tyenv)
        # Need to actually implement this later
        return MultiValues(self.state.top(bits))

    @binop_handler
    def _handle_binop_16HLto32(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        return expr0.concat(expr1)

    @binop_handler
    def _handle_binop_Add(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is not None and expr1_v is not None:
            # adding two single values together
            r = MultiValues(expr0_v + expr1_v)
        elif expr0_v is None and expr1_v is not None:
            # adding a single value to a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {v.sign_extend(expr1_v.size() - v.size()) + expr1_v for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # adding a single value to a multivalue
            if expr1.count() == 1 and 0 in expr1:
                vs = {expr0_v + v.sign_extend(expr0_v.size() - v.size()) for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # we do not support addition between two real multivalues
            r = MultiValues(self.state.top(bits))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Sub(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is not None and expr1_v is not None:
            # subtracting a single value from another single value
            r = MultiValues(expr0_v - expr1_v)
        elif expr0_v is None and expr1_v is not None:
            # subtracting a single value from a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {v - expr1_v for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # subtracting a single value from a multivalue
            if expr1.count() == 1 and 0 in expr1:
                vs = {expr0_v - v for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # we do not support addition between two real multivalues
            r = MultiValues(self.state.top(bits))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Mul(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_pair(expr.args[0], expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            # we do not support multiplication between two real multivalues
            r = MultiValues(self.state.top(bits))
        elif expr0_v is None and expr1_v is not None:
            # multiplying a single value to a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {v * expr1_v for v in expr0[0]}  # type: ignore
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # multiplying a single value to a multivalue
            if expr1.count() == 1 and 0 in expr1:
                vs = {v * expr0_v for v in expr1[0]}  # type: ignore
                r = MultiValues(offset_to_values={0: vs})
        else:
            # multiplying two single values together
            r = MultiValues(expr0_v * expr1_v)  # type: ignore

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Mull(self, expr: pyvex.expr.Binop) -> MultiValues:
        _, _ = self._expr(expr.args[0]), self._expr(expr.args[1])
        bits = expr.result_size(self.tyenv)
        return MultiValues(self.state.top(bits))

    @binop_handler
    def _handle_binop_Div(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_pair(expr.args[0], expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is not None and expr1_v is not None:
            if expr0_v.concrete and expr1_v.concrete:
                # dividing two single values
                r = MultiValues(self.state.top(bits)) if expr1_v.concrete_value == 0 else MultiValues(expr0_v / expr1_v)  # type: ignore
        elif expr0_v is None and expr1_v is not None:
            if expr1_v.concrete and expr1_v.concrete_value == 0:
                r = MultiValues(self.state.top(bits))
            elif expr0.count() == 1 and 0 in expr0:
                vs = {v / expr1_v for v in expr0[0]}  # type: ignore
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            if expr1.count() == 1 and 0 in expr1:
                vs = {expr0_v / v for v in expr1[0] if (not v.concrete) or v.concrete_value != 0}  # type: ignore
                r = MultiValues(offset_to_values={0: vs})
        else:
            # we do not support division between two real multivalues
            r = MultiValues(self.state.top(bits))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_DivMod(self, expr: pyvex.expr.Binop) -> MultiValues:
        _, _ = self._expr(expr.args[0]), self._expr(expr.args[1])
        bits = expr.result_size(self.tyenv)

        return MultiValues(self.state.top(bits))

    @binop_handler
    def _handle_Mod(self, expr: pyvex.expr.Binop) -> MultiValues:
        _, _ = self._expr(expr.args[0]), self._expr(expr.args[1])
        bits = expr.result_size(self.tyenv)
        return MultiValues(self.state.top(bits))

    @binop_handler
    def _handle_binop_And(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is not None and expr1_v is not None:
            # bitwise-and two single values together
            r = MultiValues(expr0_v & expr1_v)
        elif expr0_v is None and expr1_v is not None:
            # bitwise-and a single value with a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {v & expr1_v for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # bitwise-and a single value to a multivalue
            if expr1.count() == 1 and 0 in expr1:
                vs = {v & expr0_v for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # we do not support addition between two real multivalues
            r = MultiValues(self.state.top(bits))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Xor(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is not None and expr1_v is not None:
            if expr0_v.concrete and expr1_v.concrete:
                # bitwise-xor two single values together
                r = MultiValues(expr0_v ^ expr1_v)
        elif expr0_v is None and expr1_v is not None:
            # bitwise-xor a single value with a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {v.sign_extend(expr1_v.size() - v.size()) ^ expr1_v for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # bitwise-xor a single value to a multivalue
            if expr1.count() == 1 and 0 in expr1:
                vs = {v.sign_extend(expr0_v.size() - v.size()) ^ expr0_v for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # we do not support xor between two real multivalues
            r = MultiValues(self.state.top(bits))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Or(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is not None and expr1_v is not None:
            # bitwise-and two single values together
            r = MultiValues(expr0_v | expr1_v)
        elif expr0_v is None and expr1_v is not None:
            # bitwise-or a single value with a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {v | expr1_v for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # bitwise-or a single value to a multivalue
            if expr1.count() == 1 and 0 in expr1:
                vs = {v | expr0_v for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # we do not support or between two real multivalues
            r = MultiValues(self.state.top(bits))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Sar(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        def _shift_sar(e0: claripy.ast.BV, e1: claripy.ast.BV):
            # convert e1 to an integer to prevent claripy from complaining "args' lengths must all be equal"
            if e1.symbolic:
                return self.state.top(bits)
            e1_int = e1.concrete_value

            if e1_int > bits:
                return claripy.BVV(0, bits)

            head = claripy.BVV(0, bits) if claripy.is_true(e0 >> bits - 1 == 0) else (1 << e1_int) - 1 << bits - e1_int
            return head | (e0 >> e1_int)

        if expr0_v is not None and expr1_v is not None:
            # subtracting a single value from another single value
            r = MultiValues(_shift_sar(expr0_v, expr1_v))
        elif expr0_v is None and expr1_v is not None:
            # shifting a single value by a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {_shift_sar(v, expr1_v) for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # shifting a multivalue by a single value
            if expr1.count() == 1 and 0 in expr1:
                vs = {_shift_sar(expr0_v, v) for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # we do not support shifting between two real multivalues
            r = MultiValues(self.state.top(bits))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Shr(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr_bv(expr.args[0]), self._expr_bv(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        def _shift_shr(e0, e1):
            if e1.symbolic:
                return self.state.top(bits)
            if e1.size() < e0.size():
                e1 = e1.sign_extend(e0.size() - e1.size())
            else:
                e0 = e0.sign_extend(e1.size() - e0.size())

            return claripy.LShR(e0, e1)

        if expr0_v is None and expr1_v is None:
            # we do not support shifting between two real multivalues
            r = MultiValues(self.state.top(bits))
        elif expr0_v is None and expr1_v is not None:
            # shifting a single value by a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {_shift_shr(v, expr1_v) for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # shifting a multivalue by a single value
            if expr1.count() == 1 and 0 in expr1:
                vs = {_shift_shr(expr0_v, v) for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # shifting a single value from another single value
            r = MultiValues(_shift_shr(expr0_v, expr1_v))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_Shl(self, expr: pyvex.expr.Binop) -> MultiValues:
        expr0, expr1 = self._expr(expr.args[0]), self._expr(expr.args[1])
        bits = expr.result_size(self.tyenv)

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        def _shift_shl(e0, e1):
            # convert e1 to an integer to prevent claripy from complaining "args' lengths must all be equal"
            if e1.symbolic:
                return self.state.top(bits)
            e1 = e1.concrete_value
            return e0 << e1

        if expr0_v is None and expr1_v is None:
            # we do not support shifting between two real multivalues
            r = MultiValues(self.state.top(bits))
        elif expr0_v is None and expr1_v is not None:
            # shifting left a single value by a multivalue
            if expr0.count() == 1 and 0 in expr0:
                vs = {_shift_shl(v, expr1_v) for v in expr0[0]}
                r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # shifting left a multivalue by a single value
            if expr1.count() == 1 and 0 in expr1:
                vs = {_shift_shl(expr0_v, v) for v in expr1[0]}
                r = MultiValues(offset_to_values={0: vs})
        else:
            # subtracting a single value from another single value
            r = MultiValues(_shift_shl(expr0_v, expr1_v))

        if r is None:
            r = MultiValues(self.state.top(bits))

        return r

    @binop_handler
    def _handle_binop_CmpEQ(self, expr: pyvex.expr.Binop) -> MultiValues:
        arg0, arg1 = expr.args
        expr_0 = self._expr(arg0)
        expr_1 = self._expr(arg1)

        e0 = expr_0.one_value()
        e1 = expr_1.one_value()

        if e0 is not None and e1 is not None:
            if not e0.symbolic and not e1.symbolic:
                return MultiValues(claripy.BVV(1, 1) if e0.concrete_value == e1.concrete_value else claripy.BVV(0, 1))
            if e0 is e1:
                return MultiValues(claripy.BVV(1, 1))
            return MultiValues(self.state.top(1))

        return MultiValues(self.state.top(1))

    @binop_handler
    def _handle_binop_CmpNE(self, expr: pyvex.expr.Binop) -> MultiValues:
        arg0, arg1 = expr.args
        expr_0 = self._expr(arg0)
        expr_1 = self._expr(arg1)

        e0 = expr_0.one_value()
        e1 = expr_1.one_value()
        if e0 is not None and e1 is not None:
            if not e0.symbolic and not e1.symbolic:
                return MultiValues(claripy.BVV(1, 1) if e0.concrete_value != e1.concrete_value else claripy.BVV(0, 1))
            if e0 is e1:
                return MultiValues(claripy.BVV(0, 1))
        return MultiValues(self.state.top(1))

    @binop_handler
    def _handle_binop_CmpLT(self, expr: pyvex.expr.Binop) -> MultiValues:
        arg0, arg1 = expr.args
        expr_0, expr_1 = self._expr_pair(arg0, arg1)

        e0 = expr_0.one_value()
        e1 = expr_1.one_value()
        if e0 is not None and e1 is not None:
            if not e0.symbolic and not e1.symbolic:
                cmp = e0.concrete_value < e1.concrete_value  # type: ignore
                return MultiValues(claripy.BVV(1, 1) if cmp else claripy.BVV(0, 1))
            if e0 is e1:
                return MultiValues(claripy.BVV(0, 1))
        return MultiValues(self.state.top(1))

    @binop_handler
    def _handle_binop_CmpLE(self, expr: pyvex.expr.Binop) -> MultiValues:
        arg0, arg1 = expr.args
        expr_0, expr_1 = self._expr_pair(arg0, arg1)

        e0 = expr_0.one_value()
        e1 = expr_1.one_value()
        if e0 is not None and e1 is not None:
            if not e0.symbolic and not e1.symbolic:
                cmp = e0.concrete_value <= e1.concrete_value  # type: ignore
                return MultiValues(claripy.BVV(1, 1) if cmp else claripy.BVV(0, 1))
            if e0 is e1:
                return MultiValues(claripy.BVV(0, 1))
        return MultiValues(self.state.top(1))

    @binop_handler
    def _handle_binop_CmpGT(self, expr: pyvex.expr.Binop) -> MultiValues:
        arg0, arg1 = expr.args
        expr_0, expr_1 = self._expr_pair(arg0, arg1)

        e0 = expr_0.one_value()
        e1 = expr_1.one_value()
        if e0 is not None and e1 is not None:
            if not e0.symbolic and not e1.symbolic:
                cmp = e0.concrete_value > e1.concrete_value  # type: ignore
                return MultiValues(claripy.BVV(1, 1) if cmp else claripy.BVV(0, 1))
            if e0 is e1:
                return MultiValues(claripy.BVV(0, 1))
        return MultiValues(self.state.top(1))

    @binop_handler
    def _handle_binop_CmpGE(self, expr: pyvex.expr.Binop) -> MultiValues:
        arg0, arg1 = expr.args
        expr_0, expr_1 = self._expr_pair(arg0, arg1)

        e0 = expr_0.one_value()
        e1 = expr_1.one_value()
        if e0 is not None and e1 is not None:
            if not e0.symbolic and not e1.symbolic:
                cmp = e0.concrete_value >= e1.concrete_value  # type: ignore
                return MultiValues(claripy.BVV(1, 1) if cmp else claripy.BVV(0, 1))
            if e0 is e1:
                return MultiValues(claripy.BVV(0, 1))
        return MultiValues(self.state.top(1))

    # ppc only
    @binop_handler
    def _handle_binop_CmpORD(self, expr: pyvex.expr.Binop) -> MultiValues:
        arg0, arg1 = expr.args
        expr_0, expr_1 = self._expr_pair(arg0, arg1)

        e0 = expr_0.one_value()
        e1 = expr_1.one_value()
        bits = expr.result_size(self.tyenv)

        if e0 is not None and e1 is not None:
            if not e0.symbolic and not e1.symbolic:
                e0 = e0.concrete_value
                e1 = e1.concrete_value
                if e0 < e1:  # type: ignore
                    return MultiValues(claripy.BVV(0x8, bits))
                if e0 > e1:  # type: ignore
                    return MultiValues(claripy.BVV(0x4, bits))
                return MultiValues(claripy.BVV(0x2, bits))
            if e0 is e1:
                return MultiValues(claripy.BVV(0x2, bits))

        return MultiValues(self.state.top(bits))

    def _handle_expr_CCall(self, expr) -> MultiValues[claripy.ast.BV | claripy.ast.FP]:
        bits = expr.result_size(self.tyenv)
        for arg_expr in expr.args:
            self._expr(arg_expr)
        return MultiValues(self.state.top(bits))

    def _handle_expr_GSPTR(self, expr):
        return self._top(expr.result_size(self.tyenv))

    def _handle_expr_VECRET(self, expr):
        return self._top(expr.result_size(self.tyenv))

    #
    # User defined high level statement handlers
    #

    def _handle_function(self, func_addr: MultiValues[claripy.ast.BV] | None):
        if func_addr is None:
            func_addr = MultiValues(self.state.top(self.state.arch.bits))

        callsite = self.state.codeloc
        data = FunctionCallData(
            callsite,
            self._function_handler.make_function_codeloc(func_addr, callsite, self.state.analysis.model.func_addr),
            func_addr,
            visited_blocks=set(),
        )
        self._function_handler.handle_function(self.state, data)
        self._visited_blocks = data.visited_blocks

    def _tag_definitions_of_atom(self, atom: Atom, func_addr: int):
        definitions = self.state.get_definitions(atom)
        tag = ParameterTag(function=func_addr, metadata={"tagged_by": "SimEngineRDVEX._handle_function_cc"})
        for definition in definitions:
            definition.tags |= {tag}



from __future__ import annotations

import logging
import struct

import claripy

from angr.sim_state import SimState
from angr.storage.memory_mixins import PagedMemoryMixin
from .plugin import SimStatePlugin

l = logging.getLogger(name=__name__)


def _mem_write_cb(s):
    s.symbolizer._mem_write_callback()


def _mem_read_cb(s):
    s.symbolizer._mem_read_callback()


def _reg_write_cb(s):
    s.symbolizer._reg_write_callback()


def _reg_read_cb(s):
    s.symbolizer._reg_read_callback()


def _page_map_cb(s):
    s.symbolizer._page_map_callback()


PAGE_SIZE = 0x1000


class SimSymbolizer(SimStatePlugin):  # pylint:disable=abstract-method
    """
    The symbolizer state plugin ensures that pointers that are stored in memory are symbolic.
    This allows for the tracking of and reasoning over these pointers (for example, to reason
    about memory disclosure).
    """

    def __init__(self):
        SimStatePlugin.__init__(self)

        self._symbolize_all = False
        self.symbolization_target_pages = set()
        self.ignore_target_pages = set()
        self.symbolized_count = 0
        self.page_symbols = {}
        self._min_addr = 0
        self._max_addr = 0

        self._LE_FMT = None
        self._BE_FMT = None
        self._zero = None

    def _page_map_callback(self):
        if self._symbolize_all:
            page_id, _ = self.state.memory._divide_addr(self.state.inspect.mapped_address)
            self.symbolization_target_pages.add(page_id)

    def _mem_write_callback(self):
        if not isinstance(self.state.inspect.mem_write_expr, int) and self.state.inspect.mem_write_expr.symbolic:
            return
        mem_write_length = self.state.inspect.mem_write_length
        if mem_write_length is not None and not isinstance(mem_write_length, int) and mem_write_length.symbolic:
            return

        # length = self.state.solver.eval_one(self.state.inspect.mem_write_length)
        # if length != self.state.arch.bytes:
        #   return

        write_expr = self.state.inspect.mem_write_expr
        byte_expr = self.state.solver.eval_one(write_expr, cast_to=bytes).rjust(
            write_expr.length // self.state.arch.byte_width
        )
        replacement_expr = self._resymbolize_data(byte_expr)
        if replacement_expr is not None:
            assert replacement_expr.length == write_expr.length
            self.state.inspect.mem_write_expr = replacement_expr

    def _reg_write_callback(self):
        if not isinstance(self.state.inspect.reg_write_expr, int) and self.state.inspect.reg_write_expr.symbolic:
            return
        if not isinstance(self.state.inspect.reg_write_length, int) and self.state.inspect.reg_write_length.symbolic:
            return
        if self.state.inspect.reg_write_offset == self.state.arch.ip_offset:
            return

        length = self.state.solver.eval_one(self.state.inspect.reg_write_length)
        if length != self.state.arch.bytes:
            return

        expr = self.state.solver.eval_one(self.state.inspect.reg_write_expr)
        if self._should_symbolize(expr):
            self.state.inspect.reg_write_expr = self._preconstrain(expr)

    def init_state(self):
        super().init_state()

        if not isinstance(self.state.memory, PagedMemoryMixin):
            raise TypeError("Symbolizer only supports a paged memory model.")

        assert self.state.memory.page_size == PAGE_SIZE

        self._LE_FMT = self.state.arch.struct_fmt(endness="Iend_LE")
        self._BE_FMT = self.state.arch.struct_fmt(endness="Iend_BE")

        # ignore CLE pages
        for i in range(0, self.state.project.loader.kernel_object.map_size, PAGE_SIZE):
            self.ignore_target_pages.add((self.state.project.loader.kernel_object.mapped_base + i) // PAGE_SIZE)
        for i in range(0, self.state.project.loader.extern_object.map_size, PAGE_SIZE):
            self.ignore_target_pages.add((self.state.project.loader.extern_object.mapped_base + i) // PAGE_SIZE)

        self.state.inspect.make_breakpoint("memory_page_map", when=self.state.inspect.BP_BEFORE, action=_page_map_cb)
        self.state.inspect.make_breakpoint("mem_write", when=self.state.inspect.BP_BEFORE, action=_mem_write_cb)
        # self.state.inspect.make_breakpoint('mem_read', when=self.state.inspect.BP_BEFORE, action=_mem_read_cb)
        # self.state.inspect.make_breakpoint('reg_write', when=self.state.inspect.BP_BEFORE, action=_reg_write_cb)
        # self.state.inspect.make_breakpoint('reg_read', when=self.state.inspect.BP_BEFORE, action=_reg_read_cb)

        self._zero = claripy.BVV(0, self.state.arch.bytes)

    @staticmethod
    def _page_id(x):
        return x // PAGE_SIZE

    @staticmethod
    def _page_addr(p):
        return p * PAGE_SIZE

    @staticmethod
    def _page_offset(p):
        return p % PAGE_SIZE

    def _update_ranges(self):
        if len(self.symbolization_target_pages) == 0:
            self._max_addr = self._min_addr = 0
        else:
            self._min_addr = self._page_addr(min(self.symbolization_target_pages))
            self._max_addr = self._page_addr(max(self.symbolization_target_pages) + 1)

    def set_symbolization_for_all_pages(self):
        """
        Sets the symbolizer to symbolize pointers to all pages as they are written to memory..
        """
        self._symbolize_all = True
        self.symbolization_target_pages.update(set(self.state.memory._pages.keys()))
        self._update_ranges()

    def set_symbolized_target_range(self, base, length):
        """
        All pointers to the target range will be symbolized as they are written to memory.

        Due to optimizations, the _pages_ containing this range will be set as symbolization targets,
        not just the range itself.
        """
        base_page = self._page_id(base)
        pages = (length + self._page_offset(base) + PAGE_SIZE - 1) // PAGE_SIZE
        assert pages > 0
        self.symbolization_target_pages.update(range(base_page, base_page + pages))
        self._update_ranges()

    def _preconstrain(self, value, name_prefix="address_"):
        page_base = value & ~(PAGE_SIZE - 1)
        try:
            symbol = self.page_symbols[page_base]
        except KeyError:
            symbol = claripy.BVS(name_prefix + hex(page_base), self.state.arch.bits)
            self.page_symbols[page_base] = symbol
            self.state.add_constraints(symbol == page_base)
        self.symbolized_count += 1
        return symbol + (value - page_base)

    def _should_symbolize(self, addr):
        return (
            self._page_id(addr) in self.symbolization_target_pages
            and self._page_id(addr) not in self.ignore_target_pages
        )

    def _resymbolize_int(self, be, le=0, base=0, offset=0, skip=()):
        if base + offset in skip:
            return None
        if self._min_addr <= be < self._max_addr and self._should_symbolize(be):
            s = self._preconstrain(be)
            l.debug("Replacing %#x (at %#x, endness BE) with %s!", be, base + offset, s)
            return s
        if self._min_addr <= le < self._max_addr and self._should_symbolize(le):
            s = self._preconstrain(le).reversed
            l.debug("Replacing %#x (at %#x, endness LE) with %s!", le, base + offset, s)
            return s
        return None

    def _resymbolize_data(self, data, prefix=b"", base=0, skip=()):
        ws = self.state.arch.bytes
        suffix = data[len(data) - (len(data) % ws) :]
        data = data[: len(data) - (len(data) % ws)]

        num_words = len(data) // ws
        unpacked_le = struct.unpack(self._LE_FMT[0] + str(num_words) + self._LE_FMT[1], data)
        unpacked_be = struct.unpack(self._BE_FMT[0] + str(num_words) + self._BE_FMT[1], data)

        values_squashed = [prefix]
        last_idx = 0
        for i, (be, le) in enumerate(zip(unpacked_be, unpacked_le)):
            # assert len(claripy.Concat(*values_squashed)) == i*8

            s = self._resymbolize_int(be, le, base, i * ws, skip)
            if s is None:
                return None

            if last_idx != i:
                values_squashed.append(data[last_idx * ws : i * ws])
            last_idx = i + 1
            values_squashed.append(s)

        if len(values_squashed) == 1:
            return None

        if last_idx != num_words:
            values_squashed.append(data[last_idx * ws :])
        values_squashed.append(suffix)

        return claripy.Concat(*values_squashed)
        # assert len(new_data)/8 == len(data) + len(prefix)
        # assert self.state.solver.eval_one(new_data) == self.state.solver.eval_one(claripy.BVV(data))

    def _resymbolize_region(self, storage: PagedMemoryMixin, addr, length):
        assert type(addr) is int
        assert type(length) is int

        self.state.scratch.push_priv(True)
        i = 0
        while i < length:
            a = addr + i
            aligned_a = a - a % (-self.state.arch.bytes)
            i = aligned_a - addr
            if i == length:
                break

            data = storage.concrete_load(aligned_a, length - i)
            if not data or len(data) == 0:
                # try the next byte
                i += 1
                continue
            replacement_content = self._resymbolize_data(
                bytearray(data),
                base=aligned_a,
                prefix=b"",
                skip=() if storage is self.state.memory else (self.state.arch.ip_offset,),
            )
            if replacement_content is not None:
                storage.store(aligned_a, replacement_content)
                i += len(replacement_content) // self.state.arch.bytes
            else:
                i += 1

        self.state.scratch.pop_priv()

    def resymbolize(self):
        """
        Re-symbolizes all pointers in memory. This can be called to symbolize any pointers to target regions
        that were written (and not mangled beyond recognition) before symbolization was set.
        """

        for i, p_id in enumerate(self.state.memory._pages):
            if i % 100 == 0:
                l.info("%s/%s memory pages symbolized", i, len(self.state.memory._pages))
            addr_start = p_id * self.state.memory.page_size
            length = self.state.memory.page_size
            self._resymbolize_region(self.state.memory, addr_start, length)

    @SimStatePlugin.memo
    def copy(self, memo):  # pylint: disable=unused-argument
        sc = SimSymbolizer()
        sc._symbolize_all = self._symbolize_all
        sc.symbolization_target_pages = set(self.symbolization_target_pages)
        sc.ignore_target_pages = set(self.ignore_target_pages)
        sc.symbolized_count = self.symbolized_count
        sc._LE_FMT = self._LE_FMT
        sc._BE_FMT = self._BE_FMT
        sc._min_addr = self._min_addr
        sc._max_addr = self._max_addr
        sc.page_symbols = dict(sc.page_symbols)
        return sc


SimState.register_default("symbolizer", SimSymbolizer)



from __future__ import annotations
import traceback
import itertools

event_id_count = itertools.count()


class SimEvent:
    """
    A SimEvent is a log entry for some notable event during symbolic execution. It logs the location it was generated
    (ins_addr, bbl_addr, stmt_idx, and sim_procedure) as well as arbitrary tags (objects).

    You may also be interested in SimAction, which is a specialization of SimEvent for CPU events.
    """

    # def __init__(self, address=None, stmt_idx=None, message=None, exception=None, traceback=None):
    def __init__(self, state, event_type, **kwargs):
        self.id = next(event_id_count)
        self.type = event_type
        self.ins_addr = state.scratch.ins_addr
        self.bbl_addr = state.scratch.bbl_addr
        self.stmt_idx = state.scratch.stmt_idx
        self.sim_procedure = None if state.scratch.sim_procedure is None else state.scratch.sim_procedure.canonical
        self.objects = dict(kwargs)
        self.arch = state.arch

    def __repr__(self):
        return f"<SimEvent {self.type} {self.id}, with fields {', '.join(self.objects.keys())}>"

    def _copy_event(self):
        c = self.__class__.__new__(self.__class__)
        c.id = self.id
        c.type = self.type
        c.bbl_addr = self.bbl_addr
        c.stmt_idx = self.stmt_idx
        c.sim_procedure = self.sim_procedure
        c.objects = dict(self.objects)

        return c


def resource_event(state, exception):
    for frame, lineno in reversed(list(traceback.walk_tb(exception.__traceback__))):
        module = frame.f_globals.get("__name__", "").split(".")
        function = frame.f_code.co_name
        if module[0] == "claripy" or module in (
            ["angr", "state_plugins", "solver"],
            ["angr", "state_plugins", "sim_action_object"],
        ):
            continue
        state.history.add_event(
            "insufficient_resources",
            module=module,
            function=function,
            lineno=lineno,
            type=type(exception),
            reason=exception.args,
        )
        break



from __future__ import annotations
import claripy

import angr


class ispunct(angr.SimProcedure):
    # pylint: disable=arguments-differ, missing-class-docstring
    def run(self, c):
        is_punct_1 = claripy.And(c >= 33, c <= 47)  # ! to /
        is_punct_2 = claripy.And(c >= 58, c <= 64)  # : to @
        is_punct_3 = claripy.And(c >= 91, c <= 96)  # [ to `
        is_punct_4 = claripy.And(c >= 123, c <= 126)  # { to ~
        int_size = self.arch.sizeof["int"]  # type: ignore[reportOptionalMemberAccess]

        return claripy.If(
            claripy.Or(is_punct_1, is_punct_2, is_punct_3, is_punct_4),
            claripy.BVV(1, int_size),
            claripy.BVV(0, int_size),
        )



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: angr/protos/cfg.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'angr/protos/cfg.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from angr.protos import primitives_pb2 as angr_dot_protos_dot_primitives__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x61ngr/protos/cfg.proto\x12\x0b\x61ngr.protos\x1a\x1c\x61ngr/protos/primitives.proto\"1\n\rOptionalInt64\x12\x11\n\thas_value\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\x03\"d\n\x0c\x42lockIDProto\x12\x0c\n\x04\x61\x64\x64r\x18\x01 \x01(\x04\x12\x33\n\x0f\x63\x61llsite_tuples\x18\x02 \x03(\x0b\x32\x1a.angr.protos.OptionalInt64\x12\x11\n\tjump_type\x18\x03 \x01(\t\"_\n\x0b\x43\x46GEdgeData\x12,\n\x08jumpkind\x18\x01 \x01(\x0e\x32\x1a.angr.protos.Edge.JumpKind\x12\x10\n\x08ins_addr\x18\x02 \x01(\x04\x12\x10\n\x08stmt_idx\x18\x03 \x01(\x05\"\xd2\x02\n\x07\x43\x46GNode\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\r\x12\x10\n\x08\x62lock_id\x18\x03 \x03(\x04\x12\x1e\n\x11simprocedure_name\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x06no_ret\x18\x07 \x01(\x08H\x01\x88\x01\x01\x12\x1d\n\x10\x66unction_address\x18\x08 \x01(\x04H\x02\x88\x01\x01\x12\r\n\x05thumb\x18\t \x01(\x08\x12\x18\n\x0b\x62yte_string\x18\n \x01(\x0cH\x03\x88\x01\x01\x12\x11\n\x04name\x18\x0b \x01(\tH\x04\x88\x01\x01\x12\x12\n\nis_syscall\x18\x0c \x01(\x08\x12\x15\n\rins_base_addr\x18\r \x01(\x04\x12\x11\n\tins_sizes\x18\x0e \x01(\x0c\x42\x14\n\x12_simprocedure_nameB\t\n\x07_no_retB\x13\n\x11_function_addressB\x0e\n\x0c_byte_stringB\x07\n\x05_name\"\xf5\x02\n\x08\x43\x46GENode\x12\"\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x14.angr.protos.CFGNode\x12\x31\n\rcallstack_key\x18\x02 \x03(\x0b\x32\x1a.angr.protos.OptionalInt64\x12\x19\n\x0csyscall_name\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x15\n\rlooping_times\x18\x04 \x01(\x05\x12\x12\n\x05\x64\x65pth\x18\x05 \x01(\x05H\x01\x88\x01\x01\x12\x1a\n\rreturn_target\x18\x06 \x01(\x04H\x02\x88\x01\x01\x12\"\n\x15\x63reation_failure_info\x18\x07 \x01(\tH\x03\x88\x01\x01\x12\x34\n\x0c\x62lock_id_obj\x18\x08 \x01(\x0b\x32\x19.angr.protos.BlockIDProtoH\x04\x88\x01\x01\x42\x0f\n\r_syscall_nameB\x08\n\x06_depthB\x10\n\x0e_return_targetB\x18\n\x16_creation_failure_infoB\x0f\n\r_block_id_obj\"\xee\x01\n\x03\x43\x46G\x12\r\n\x05ident\x18\x01 \x01(\t\x12#\n\x05nodes\x18\x02 \x03(\x0b\x32\x14.angr.protos.CFGNode\x12 \n\x05\x65\x64ges\x18\x03 \x03(\x0b\x32\x11.angr.protos.Edge\x12,\n\x0bmemory_data\x18\x04 \x03(\x0b\x32\x17.angr.protos.MemoryData\x12\x12\n\nnormalized\x18\x05 \x01(\x08\x12.\n\x0bjump_tables\x18\x06 \x03(\x0b\x32\x19.angr.protos.IndirectJump\x12\x1f\n\x17\x62lock_addrs_with_return\x18\x07 \x03(\x04\"\xc5\x03\n\nMemoryData\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x11\n\x04size\x18\x02 \x01(\rH\x00\x88\x01\x01\x12\x34\n\x04type\x18\x03 \x01(\x0e\x32&.angr.protos.MemoryData.MemoryDataType\x12\x1b\n\x0ereference_size\x18\x04 \x01(\rH\x01\x88\x01\x01\"\xa8\x02\n\x0eMemoryDataType\x12\x13\n\x0fUnknownDataType\x10\x00\x12\x0f\n\x0bUnspecified\x10\x01\x12\x0b\n\x07Integer\x10\x02\x12\x10\n\x0cPointerArray\x10\x03\x12\n\n\x06String\x10\x04\x12\x11\n\rUnicodeString\x10\x05\x12\x13\n\x0fSegmentBoundary\x10\x06\x12\x11\n\rCodeReference\x10\x07\x12\x0f\n\x0bGOTPLTEntry\x10\x08\x12\r\n\tELFHeader\x10\t\x12\x11\n\rFloatingPoint\x10\n\x12\r\n\tAlignment\x10\x0b\x12\x15\n\x11PEImportDirectory\x10\x0c\x12\x15\n\x11PEExportDirectory\x10\r\x12\x1a\n\x16PEDelayImportDirectory\x10\x0e\x42\x07\n\x05_sizeB\x11\n\x0f_reference_size\"^\n\rJumptableInfo\x12\x11\n\x04\x61\x64\x64r\x18\x01 \x01(\x04H\x00\x88\x01\x01\x12\x0c\n\x04size\x18\x02 \x01(\x05\x12\x12\n\nentry_size\x18\x03 \x01(\x05\x12\x0f\n\x07\x65ntries\x18\x04 \x03(\x04\x42\x07\n\x05_addr\"\xd0\x01\n\x0cIndirectJump\x12\x0c\n\x04\x61\x64\x64r\x18\x01 \x01(\x04\x12\x10\n\x08ins_addr\x18\x02 \x01(\x04\x12\x11\n\tfunc_addr\x18\x03 \x01(\x04\x12\x10\n\x08jumpkind\x18\x04 \x01(\t\x12\x10\n\x08stmt_idx\x18\x05 \x01(\x05\x12\x18\n\x10resolved_targets\x18\x06 \x03(\x04\x12\x11\n\tjumptable\x18\x07 \x01(\x08\x12.\n\njumptables\x18\x08 \x03(\x0b\x32\x1a.angr.protos.JumptableInfo\x12\x0c\n\x04type\x18\t \x01(\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'angr.protos.cfg_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_OPTIONALINT64']._serialized_start=68
  _globals['_OPTIONALINT64']._serialized_end=117
  _globals['_BLOCKIDPROTO']._serialized_start=119
  _globals['_BLOCKIDPROTO']._serialized_end=219
  _globals['_CFGEDGEDATA']._serialized_start=221
  _globals['_CFGEDGEDATA']._serialized_end=316
  _globals['_CFGNODE']._serialized_start=319
  _globals['_CFGNODE']._serialized_end=657
  _globals['_CFGENODE']._serialized_start=660
  _globals['_CFGENODE']._serialized_end=1033
  _globals['_CFG']._serialized_start=1036
  _globals['_CFG']._serialized_end=1274
  _globals['_MEMORYDATA']._serialized_start=1277
  _globals['_MEMORYDATA']._serialized_end=1730
  _globals['_MEMORYDATA_MEMORYDATATYPE']._serialized_start=1406
  _globals['_MEMORYDATA_MEMORYDATATYPE']._serialized_end=1702
  _globals['_JUMPTABLEINFO']._serialized_start=1732
  _globals['_JUMPTABLEINFO']._serialized_end=1826
  _globals['_INDIRECTJUMP']._serialized_start=1829
  _globals['_INDIRECTJUMP']._serialized_end=2037
# @@protoc_insertion_point(module_scope)
